"""
This program contains a function to compute multi-infeed admittance matrices both for AC and DC systems

Copyright (C) 2024  Francisco Javier Cifuentes Garcia

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['admittance','admittance_multi_freq','admittance_generic']

import matplotlib.pyplot as plt
import numpy as np  # Numerical python functions
from matplotlib import rcParams  # Text's parameters for plots

rcParams['mathtext.fontset'] = 'cm'  # Font selection
rcParams['font.family'] = 'STIXGeneral'  # 'cmu serif'

def admittance(f_base=None, frequencies=None, fft_periods=1, scantype="AC", sides=None, dt=None, exploit_dq_sym=False,
               start_idx=None, zblocks=None, results_folder=None, results_name='Y', network=None):
    # Small-signal sinusoidal steady state computation and rFFT (no target frequency-based FFT distinction)
    L = int(fft_periods * 1 / f_base * 1.0 / dt)  # For the FFT computation
    if scantype is "AC":
        # Compute the small-signal sinusoidal steady-state waveforms
        # For each simulation (freq) it contains a matrix: [Vd_d Vd_q; Vq_d Vq_q]
        deltaV = np.empty((len(frequencies), 2, 2), dtype='cdouble')  # Also dtype='csingle'
        deltaI = np.empty((len(frequencies), 2, 2), dtype='cdouble')  # Also dtype='csingle'
        Y = np.empty((len(frequencies), 2, 2), dtype='cdouble')  # Also dtype='csingle'
        names = ['VDUTac:1', 'VDUTac:2', 'IDUTacA' + sides + ':1', 'IDUTacA' + sides + ':2']
        row = {names[0]: 0, names[1]: 1, names[2]: 0, names[3]: 1}
        for sim, frequency in enumerate(frequencies):
            fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
            for col, sim_type in enumerate(["_d", "_q"]):
                if exploit_dq_sym:
                    if col != 1:
                        for name in names:
                            delta = zblocks.perturbation_data[sim][name+sim_type][start_idx:] - zblocks.snapshot_data[name][start_idx:]
                            delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                            # Retrieve the response at the target frequency
                            if "V" in name:
                                deltaV[sim, row[name], col] = delta_FD[fft_idx]
                            else:
                                deltaI[sim, row[name], col] = delta_FD[fft_idx]
                        
                        # The data for the q-axis is mirrored w.r.t. the d-axis data
                        deltaV[sim, row[names[0]], 1] = - deltaV[sim, row[names[1]], 0]
                        deltaV[sim, row[names[1]], 1] =   deltaV[sim, row[names[0]], 0]
                        deltaI[sim, row[names[2]], 1] = - deltaI[sim, row[names[3]], 0]
                        deltaI[sim, row[names[3]], 1] =   deltaI[sim, row[names[2]], 0]

                else:       
                    for name in names:
                        delta = zblocks.perturbation_data[sim][name+sim_type][start_idx:] - zblocks.snapshot_data[name][start_idx:]
                        delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                        # freq_FD = np.fft.rfftfreq(L, d=dt)  # rFFT frequency points
                        # Retrieve the response at the target frequency
                        if "V" in name:
                            deltaV[sim, row[name], col] = delta_FD[fft_idx]
                        else:
                            deltaI[sim, row[name], col] = delta_FD[fft_idx]
            
            Y[sim,...] = np.matmul(deltaI[sim,...], np.linalg.inv(deltaV[sim,...]))


        if results_folder is not None:
            filename = results_name + '#Y_AC#' + zblocks.name + "-" + sides
            np.savetxt(results_folder + '\\' + filename+'#.txt',
                       np.stack((frequencies, Y[:, 0, 0], Y[:, 0, 1], Y[:, 1, 0], Y[:, 1, 1]), axis=1), delimiter='\t',
                       header="f\t"+zblocks.name+"-"+sides+"_d\t"+zblocks.name+"-"+sides+"_q", comments='')
            # "f\td-d\td-q\tq-d\tq-q"
            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 0, 0])), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dd}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 0, 1])), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dq}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 1, 0])), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{qd}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 1, 1])), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{qq}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', ncol=2)

            ax[1].scatter(frequencies, np.angle(Y[:, 0, 0], deg=True), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dd}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 0, 1], deg=True), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dq}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 1, 0], deg=True), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{qd}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 1, 1], deg=True), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{qq}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            ax[1].legend(loc='upper right', fancybox=True, shadow=True, ncol=2)
            fig.savefig(results_folder + '\\' + filename + ".pdf", format="pdf", bbox_inches="tight")
            fig.clear()

    elif scantype is "DC":
        # Use the actual steady-state waveforms (allows to remove periodicity)
        # For each simulation (freq) it contains a matrix: [Vd_d Vd_q; Vq_d Vq_q]
        deltaV = np.empty((len(frequencies),), dtype='cdouble')  # Also dtype='csingle'
        deltaI = np.empty((len(frequencies),), dtype='cdouble')  # Also dtype='csingle'
        names = ['VDUTdc', 'IDUTdcA' + sides]
        for sim, frequency in enumerate(frequencies):
            fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
            for name in names:
                delta = zblocks.perturbation_data[sim][name + "_dc"][start_idx:] - zblocks.snapshot_data[name][start_idx:]
                delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                # Retrieve the response at the target frequency
                if "V" in name:
                    deltaV[sim] = delta_FD[fft_idx]
                else:
                    deltaI[sim] = delta_FD[fft_idx]
        Y = deltaI / deltaV

        if results_folder is not None:
            filename = results_name + '#Y_DC#' + zblocks.name + "-" + sides
            np.savetxt(results_folder + '\\' + filename + '#.txt', np.stack((frequencies, Y), axis=1), delimiter='\t', header="f\t"+zblocks.name+"-"+sides+"_dc", comments='')
            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y)), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', fancybox=True, shadow=True, ncol=2)

            ax[1].scatter(frequencies, np.angle(Y, deg=True), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            ax[1].legend(loc='upper right', fancybox=True, shadow=True, ncol=2)
            fig.savefig(results_folder + '\\' + filename + ".pdf", format="pdf", bbox_inches="tight")
            fig.clear()

    elif scantype is "ACDC":
        # For each simulation (freq) it contains a 3x3 matrix: [Vdc_dc Vdc_d Vdc_q; Vd_dc Vd_d Vd_q; Vq_dc Vq_d Vq_q]
        deltaV = np.empty((len(frequencies), 3, 3), dtype='cdouble')  # Also dtype='csingle'
        deltaI = np.empty((len(frequencies), 3, 3), dtype='cdouble')
        Y = np.empty((len(frequencies), 3, 3), dtype='cdouble')
        for block_num, block in enumerate(zblocks):
            if block.type == "AC":
                namesAC = ['VDUTac:1', 'VDUTac:2', 'IDUTacA' + sides[block_num] + ':1', 'IDUTacA' + sides[block_num] + ':2']
            else:
                namesDC = ['VDUTdc', 'IDUTdcA' + sides[block_num]]
        names = namesDC + namesAC
        row = {names[0]: 0, names[1]: 0, names[2]: 1, names[3]: 2, names[4]: 1, names[5]: 2}
        for sim, frequency in enumerate(frequencies):
            fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
            for col, sim_type in enumerate(["_dc", "_d", "_q"]):
                for block in zblocks:
                    # Select whether it is a DC or AC block
                    if block.type == "AC":
                        names = namesAC
                    else:
                        names = namesDC
                    for name in names:
                        delta = block.perturbation_data[sim][name + sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                        delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                        # Retrieve the response at the target frequency
                        if "V" in name:
                            deltaV[sim, row[name], col] = delta_FD[fft_idx]
                        else:
                            deltaI[sim, row[name], col] = delta_FD[fft_idx]
            Y[sim, ...] = np.matmul(deltaI[sim, ...], np.linalg.inv(deltaV[sim, ...]))

        if results_folder is not None:
            filename = results_name + '#Y_ACDC#' + zblocks[0].name+"-"+sides[0]+"#"+zblocks[1].name+"-"+sides[1]
            header = "f\tdc-dc\tdc-d\tdc-q\td-dc\td-d\td-q\tq-dc\tq-d\tq-q"
            np.savetxt(results_folder + '\\' + filename+'#.txt', np.c_[frequencies, Y.reshape(Y.shape[0],-1)], delimiter='\t',header=header,comments='')
            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 1, 1])), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dd}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 1, 2])), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dq}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 2, 1])), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{qd}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 2, 2])), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{qq}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', ncol=2)

            ax[1].scatter(frequencies, np.angle(Y[:, 1, 1], deg=True), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dd}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 1, 2], deg=True), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dq}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 2, 1], deg=True), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{qd}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 2, 2], deg=True), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{qq}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            ax[1].legend(loc='upper right', ncol=2)
            filename_aux = filename.split("#")
            filename_aux[1] = "Y_AC"
            fig.savefig(results_folder + '\\' + "#".join(filename_aux) + ".pdf", format="pdf", bbox_inches="tight")
            fig.clear()

            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 0, 0])), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', ncol=2)

            ax[1].scatter(frequencies, np.angle(Y[:, 0, 0], deg=True), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            ax[1].legend(loc='upper right', ncol=2)
            filename_aux[1] = "Y_DC"
            fig.savefig(results_folder + '\\' + "#".join(filename_aux) + ".pdf", format="pdf", bbox_inches="tight")
            fig.clear()

            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 0, 1])), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc-d}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 0, 2])), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dc-q}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 1, 0])), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{d-dc}$')
            ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:, 2, 0])), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{q-dc}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', ncol=2)

            ax[1].scatter(frequencies, np.angle(Y[:, 0, 1], deg=True), marker='o', facecolors='none', edgecolors='b',
                          linewidths=1.5, label=r'$Y_{dc-d}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 0, 2], deg=True), marker='x', c='r', linewidths=1.5,
                          label=r'$Y_{dc-q}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 1, 0], deg=True), marker='+', c='m', linewidths=1.5,
                          label=r'$Y_{d-dc}$')
            ax[1].scatter(frequencies, np.angle(Y[:, 2, 0], deg=True), marker='.', c='g', linewidths=1.5,
                          label=r'$Y_{q-dc}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            ax[1].legend(loc='upper right', ncol=2)
            filename_aux[1] = "Y_ACDC_couplings"
            fig.savefig(results_folder + '\\' + "#".join(filename_aux) + ".pdf", format="pdf", bbox_inches="tight")

    else:
        # PASSIVE NETWORK SCAN
        # Size of the admittance matrix
        if exploit_dq_sym and network.scan_type == "AC":
            N = 2*network.runs
        else:
            N = network.runs 
            
        # For each simulation (freq) a NxN matrix is computed where N = #buses for DC grids or 2 * #buses for AC grids
        deltaV = np.empty((len(frequencies), N, N), dtype='cdouble')  # Also dtype='csingle'
        deltaI = np.empty((len(frequencies), N, N), dtype='cdouble')
        Y = np.empty((len(frequencies), N, N), dtype='cdouble')

        data_ending = ["_"+str(pert) for pert in range(1,N+1)]  # Use the number of runs to define the file ending
        # The rows for each variable are based on the names of all_scans (topology), i.e. zblocks are sorted already
        # If removing the scan_type info is needed: "_".join(scan_name.split("_")[:-1])
        # row = {scan_name: idx for idx, scan_name in enumerate(network.all_scans)}

        # Build the small-signal complex voltage and current matrices at every frequency
        for sim, frequency in enumerate(frequencies):
            fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
            for col, sim_type in enumerate(data_ending):
                for block_num, block in enumerate(zblocks):
                    if network.scan_type == "AC":
                        names = ['VDUTac:1','VDUTac:2','IDUTacA'+sides[block_num]+':1','IDUTacA'+sides[block_num]+':2']
                        if exploit_dq_sym:
                            # Exploiting dq-symmetry: # runs is half and only corresponds to d-axis perturbations
                            # The data for the q-axis perturbation is derived from that of the d-axis via symmetry
                            if col < N/2:
                                # Only half of the necessary perturbation data is used (exists): d-voltage perturbed
                                for name in names:
                                    delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][ start_idx:]
                                    delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                                    row = 2 * block_num + int(name[-1]) - 1  # _d variable followed by _q variable
                                    # Retrieve the response at the target frequency
                                    if "V" in name:
                                        deltaV[sim, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis perturbation
                                    else:
                                        deltaI[sim, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis response data

                                # The data for the q-axis is mirrored w.r.t. the d-axis data
                                deltaV[sim, 2*block_num,   2*col+1] = - deltaV[sim, 2*block_num+1, 2*col]
                                deltaV[sim, 2*block_num+1, 2*col+1] = deltaV[sim, 2*block_num,   2*col]
                                deltaI[sim, 2*block_num,   2*col+1] = - deltaI[sim, 2*block_num+1, 2*col]
                                deltaI[sim, 2*block_num+1, 2*col+1] = deltaI[sim, 2*block_num,   2*col]

                        else:
                            # Computation using both d-axis and q-axis perturbations (no assumptions on the symmetry)
                            for name in names:
                                delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                                delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                                row = 2*block_num + int(name[-1]) - 1  # _d variable followed by _q variable
                                # Retrieve the response at the target frequency
                                if "V" in name:
                                    deltaV[sim, row, col] = delta_FD[fft_idx]
                                else:
                                    deltaI[sim, row, col] = delta_FD[fft_idx]
                    else:
                        names = ['VDUTdc', 'IDUTdcA' + sides[block_num]]
                        for name in names:
                            delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                            delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                            # Retrieve the response at the target frequency
                            if "V" in name:
                                deltaV[sim, block_num, col] = delta_FD[fft_idx]
                            else:
                                deltaI[sim, block_num, col] = delta_FD[fft_idx]

            if network.enforce:
                # Enforce network connectivity
                Yextended = network.adj_matrix  # Matrix indicating the connectivity
                np.fill_diagonal(Yextended, 1)  # Fill the diagonal with ones (shunt or self admittance)
                if sim == 0:
                    print("Network connectivity matrix \n",Yextended)
                if network.scan_type == "AC":
                    Yextended = np.kron(Yextended,np.ones((2,2),dtype=int))  # Extend the matrix with dq-axes
                    if sim == 0: print("\n Network dq connectivity matrix \n", Yextended)
                # Yextended[m,n] = 1 <-> y[m,n] =/= 0, so we force the rest to zero
                Y[sim, ...] = np.multiply(Yextended,np.matmul(deltaI[sim, ...], np.linalg.inv(deltaV[sim, ...])))
            else:
                Y[sim, ...] = np.matmul(deltaI[sim, ...], np.linalg.inv(deltaV[sim, ...]))

        if results_folder is not None:
            filename = results_name+'#Y_'+network.scan_type+"#"+"#".join([zblocks[idx].name+"-"+sides[idx] for idx in range(len(sides))])
            results = [Y[:, row, col] for row in range(N) for col in range(N)]
            results.insert(0, frequencies)
            elements = [str(row)+"-"+str(col) for row in range(N) for col in range(N)]
            results = tuple(results)
            header = "f\t"+"\t".join(network.all_scans)
            np.savetxt(results_folder+r'\\'+filename+'#.txt',np.stack(results, axis=1),  delimiter='\t', header=header,comments='') #, comments="\t".join(network.all_scans)

            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            for row in range(N):
                for col in range(N):
                    if network.enforce:
                        # Only plot the non-zero elements
                        if Yextended[row,col] == 1:
                            ax[0].scatter(frequencies, 20*np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                    else:
                        ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
            # ax[0].set_yscale("log")
            ax[0].set_xscale("log")
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel('Magnitude [dB]')
            ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
            ax[0].legend(loc='upper right', ncol=4)
            # ,fancybox=True, shadow=True,
            for row in range(N):
                for col in range(N):
                    if network.enforce:
                        if Yextended[row,col] == 1:
                            ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                    else:
                        ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
            ax[1].set_xscale("log")
            ax[1].set_ylim([-200, 200])
            ax[1].set_yticks([-180, -90, 0, 90, 180])
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[1].set_ylabel('Phase [°]')
            ax[1].set_xlabel('Frequency [Hz]')
            fig.savefig(results_folder+'\\'+filename + ".pdf", format="pdf", bbox_inches="tight")
            fig.clear()


def admittance_multi_freq(f_base=None, freq_multi=None, fft_periods=1, sides=None, dt=None, exploit_dq_sym=False,
                          start_idx=None, zblocks=None, results_folder=None, results_name='Y', network=None):
    # Small-signal sinusoidal steady state computation and rFFT (no target frequency-based FFT distinction)
    L = int(fft_periods * 1 / f_base * 1.0 / dt)  # For the FFT computation

    # Size of the admittance matrix
    if exploit_dq_sym and network.scan_type == "AC":
        N = 2*network.runs
    else:
        N = network.runs

    frequencies = freq_multi.reshape(np.prod(freq_multi.shape))  # Frequencies
    # TODO: remove zero entries (pad) from frequencies and freq_multi!
    
    # For each frequency a NxN matrix is computed where N = #buses for DC grids or 2 * #buses for AC grids
    deltaV = np.empty((len(frequencies), N, N), dtype='cdouble')  # Also dtype='csingle'
    deltaI = np.empty((len(frequencies), N, N), dtype='cdouble')
    Y = np.empty((len(frequencies), N, N), dtype='cdouble')

    data_ending = ["_"+str(pert) for pert in network.runs_list]  # Use the number of runs to define the file ending
    # The rows for each variable are based on the names of all_scans (topology), i.e. zblocks are sorted already
    # If removing the scan_type info is needed: "_".join(scan_name.split("_")[:-1])
    # row = {scan_name: idx for idx, scan_name in enumerate(network.all_scans)}

    # Build the small-signal complex voltage and current matrices at every frequency
    for sim in range(freq_multi.shape[1]):
        # For every simulation, read the files and extract the FFT at the corresponding frequencies
        for col, sim_type in enumerate(data_ending):
            for block_num, block in enumerate(zblocks):
                if network.scan_type == "AC":
                    names = ['VDUTac:1','VDUTac:2','IDUTacA'+sides[block_num]+':1','IDUTacA'+sides[block_num]+':2']
                    if exploit_dq_sym:
                        # Exploiting dq-symmetry: # runs is half and only corresponds to d-axis perturbations
                        # The data for the q-axis perturbation is derived from that of the d-axis via symmetry
                        if col < N/2:
                            # Only half of the necessary perturbation data is used (exists): d-voltage perturbed
                            for name in names:
                                delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][ start_idx:]
                                delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                                row = 2 * block_num + int(name[-1]) - 1  # _d variable followed by _q variable
                                for freq_file, freq in enumerate(freq_multi[..., sim]):
                                    # Retrieve the response at the target frequencies
                                    fft_idx = int(round(freq * fft_periods * 1 / f_base))  # Index of the target FFT frequency
                                    freq_idx = sim + freq_file*freq_multi.shape[1]  # Index of the current frequency in the vector of all frequencies
                                    if "V" in name:
                                        deltaV[freq_idx, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis perturbation
                                    else:
                                        deltaI[freq_idx, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis response data

                            for freq_file, freq in enumerate(freq_multi[..., sim]):
                                freq_idx = sim + freq_file*freq_multi.shape[1]  # Index of the current frequency in the vector of all frequencies
                                # The data for the q-axis is mirrored w.r.t. the d-axis data
                                deltaV[freq_idx, 2*block_num,   2*col+1] = - deltaV[freq_idx, 2*block_num+1, 2*col]
                                deltaV[freq_idx, 2*block_num+1, 2*col+1] = deltaV[freq_idx, 2*block_num,   2*col]
                                deltaI[freq_idx, 2*block_num,   2*col+1] = - deltaI[freq_idx, 2*block_num+1, 2*col]
                                deltaI[freq_idx, 2*block_num+1, 2*col+1] = deltaI[freq_idx, 2*block_num,   2*col]

                    else:
                        # Computation using both d-axis and q-axis perturbations (no assumptions on the symmetry)
                        for name in names:
                            delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                            delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                            row = 2*block_num + int(name[-1]) - 1  # _d variable followed by _q variable
                            for freq_file, freq in enumerate(freq_multi[..., sim]):
                                # Retrieve the response at the target frequencies
                                fft_idx = int(round(freq * fft_periods * 1 / f_base))  # Index of the target FFT frequency
                                freq_idx = sim + freq_file*freq_multi.shape[1]  # Index of the current frequency in the vector of all frequencies
                                if "V" in name:
                                    deltaV[freq_idx, row, col] = delta_FD[fft_idx]
                                else:
                                    deltaI[freq_idx, row, col] = delta_FD[fft_idx]
                else:
                    names = ['VDUTdc', 'IDUTdcA' + sides[block_num]]
                    for name in names:
                        delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                        delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                        for freq_file, freq in enumerate(freq_multi[..., sim]):
                            # Retrieve the response at the target frequencies
                            fft_idx = int(round(freq * fft_periods * 1 / f_base))  # Index of the target FFT frequency
                            freq_idx = sim + freq_file*freq_multi.shape[1]  # Index of the current frequency in the vector of all frequencies
                            # Retrieve the response at the target frequency
                            if "V" in name:
                                deltaV[freq_idx, block_num, col] = delta_FD[fft_idx]
                            else:
                                deltaI[freq_idx, block_num, col] = delta_FD[fft_idx]

    for freq_idx in range(len(frequencies)):
        if network.enforce:
            # Enforce network connectivity
            Yextended = network.adj_matrix  # Matrix indicating the connectivity
            np.fill_diagonal(Yextended, 1)  # Fill the diagonal with ones (shunt or self admittance)
            if freq_idx == 0:
                print("Network connectivity matrix \n",Yextended)
            if network.scan_type == "AC":
                Yextended = np.kron(Yextended,np.ones((2,2),dtype=int))  # Extend the matrix with dq-axes
                if freq_idx == 0: print("\n Network dq connectivity matrix \n", Yextended)
            # Yextended[m,n] = 1 <-> y[m,n] =/= 0, so we force the rest to zero
            Y[freq_idx, ...] = np.multiply(Yextended,np.matmul(deltaI[freq_idx, ...], np.linalg.inv(deltaV[freq_idx, ...])))
        else:
            Y[freq_idx, ...] = np.matmul(deltaI[freq_idx, ...], np.linalg.inv(deltaV[freq_idx, ...]))

    if results_folder is not None:
        filename = results_name+'#Y_'+network.scan_type+"#"+"#".join([zblocks[idx].name+"-"+sides[idx] for idx in range(len(sides))])
        results = [Y[:, row, col] for row in range(N) for col in range(N)]
        results.insert(0, frequencies)
        elements = [str(row)+"-"+str(col) for row in range(N) for col in range(N)]
        results = tuple(results)
        header = "f\t"+"\t".join(network.all_scans)
        np.savetxt(results_folder+r'\\'+filename+'#.txt',np.stack(results, axis=1),  delimiter='\t', header=header, comments='') #, comments="\t".join(network.all_scans)

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
        for row in range(N):
            for col in range(N):
                if network.enforce:
                    # Only plot the non-zero elements
                    if Yextended[row,col] == 1:
                        ax[0].scatter(frequencies, 20*np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                else:
                    ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
        # ax[0].set_yscale("log")
        ax[0].set_xscale("log")
        ax[0].set_xlim([frequencies[0], frequencies[-1]])
        ax[0].minorticks_on()
        ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[0].set_ylabel('Magnitude [dB]')
        ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
        ax[0].legend(loc='upper right', ncol=4)
        # ,fancybox=True, shadow=True,
        for row in range(N):
            for col in range(N):
                if network.enforce:
                    if Yextended[row,col] == 1:
                        ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                else:
                    ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
        ax[1].set_xscale("log")
        ax[1].set_ylim([-200, 200])
        ax[1].set_yticks([-180, -90, 0, 90, 180])
        ax[1].set_xlim([frequencies[0], frequencies[-1]])
        ax[1].minorticks_on()
        ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[1].set_ylabel('Phase [°]')
        ax[1].set_xlabel('Frequency [Hz]')
        fig.savefig(results_folder+'\\'+filename + ".pdf", format="pdf", bbox_inches="tight")
        fig.clear()


def admittance_generic(f_base=None, frequencies=None, fft_periods=1, sides=None, dt=None, exploit_dq_sym=False,
                       start_idx=None, zblocks=None, results_folder=None, results_name='Y', network=None):
    # Small-signal sinusoidal steady state computation and rFFT (no target frequency-based FFT distinction)
    L = int(fft_periods * 1 / f_base * 1.0 / dt)  # For the FFT computation
    
    if exploit_dq_sym and network.scan_type == "AC":
        N = 2*network.runs
    else:
        N = network.runs 
        
    # For each simulation (freq) a NxN matrix is computed where N = # DC buses + 2 * # AC buses
    deltaV = np.empty((len(frequencies), N, N), dtype='cdouble')  # Also dtype='csingle'
    deltaI = np.empty((len(frequencies), N, N), dtype='cdouble')
    Y = np.empty((len(frequencies), N, N), dtype='cdouble')

    data_ending = ["_"+str(run) for run in network.runs_list]  # Use the number of runs to define the file ending
    # The rows for each variable are based on the names of all_scans (topology), i.e. zblocks are sorted already
    # If removing the scan_type info is needed: "_".join(scan_name.split("_")[:-1])
    # row = {scan_name: idx for idx, scan_name in enumerate(network.all_scans)}

    # Build the small-signal complex voltage and current matrices at every frequency
    for sim, frequency in enumerate(frequencies):
        fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
        for col, sim_type in enumerate(data_ending):
            current_row = 0
            for block_num, block in enumerate(zblocks):
                if block.type == "AC":
                    names = ['VDUTac:1','VDUTac:2','IDUTacA'+sides[block_num]+':1','IDUTacA'+sides[block_num]+':2']
                    if exploit_dq_sym:
                        # Exploiting dq-symmetry: # runs is half and only corresponds to d-axis perturbations
                        # The data for the q-axis perturbation is derived from that of the d-axis via symmetry
                        if col < N/2:
                            # Only half of the necessary perturbation data is used (exists): d-voltage perturbed
                            for name in names:
                                delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][ start_idx:]
                                delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                                row = current_row + int(name[-1]) - 1  # _d variable followed by _q variable
                                # Retrieve the response at the target frequency
                                if "V" in name:
                                    deltaV[sim, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis perturbation
                                else:
                                    deltaI[sim, row, 2*col] = delta_FD[fft_idx]  # Actual d-axis response data

                            # The data for the q-axis is mirrored w.r.t. the d-axis data
                            deltaV[sim, current_row,   2*col+1] = - deltaV[sim, current_row+1, 2*col]
                            deltaV[sim, current_row+1, 2*col+1] = deltaV[sim, current_row,   2*col]
                            deltaI[sim, current_row,   2*col+1] = - deltaI[sim, current_row+1, 2*col]
                            deltaI[sim, current_row+1, 2*col+1] = deltaI[sim, current_row,   2*col]

                    else:
                        # Computation using both d-axis and q-axis perturbations (no assumptions on the symmetry)
                        for name in names:
                            delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                            delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                            row = current_row + int(name[-1]) - 1  # _d variable followed by _q variable
                            # Retrieve the response at the target frequency
                            if "V" in name:
                                deltaV[sim, row, col] = delta_FD[fft_idx]
                            else:
                                deltaI[sim, row, col] = delta_FD[fft_idx]
                    
                    current_row = current_row + 2

                else:
                    names = ['VDUTdc', 'IDUTdcA' + sides[block_num]]
                    for name in names:
                        delta = block.perturbation_data[sim][name+sim_type][start_idx:] - block.snapshot_data[name][start_idx:]
                        delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                        # Retrieve the response at the target frequency
                        if "V" in name:
                            deltaV[sim, current_row, col] = delta_FD[fft_idx]
                        else:
                            deltaI[sim, current_row, col] = delta_FD[fft_idx]
                    
                    current_row = current_row + 1

        if network.enforce:
            # Enforce network connectivity
            Yextended = network.adj_matrix  # Matrix indicating the connectivity
            np.fill_diagonal(Yextended, 1)  # Fill the diagonal with ones (shunt or self admittance)
            if sim == 0:
                print("Network connectivity matrix \n",Yextended)
            if network.scan_type == "AC":
                Yextended = np.kron(Yextended,np.ones((2,2),dtype=int))  # Extend the matrix with dq-axes
                if sim == 0: print("\n Network dq connectivity matrix \n", Yextended)
            # Yextended[m,n] = 1 <-> y[m,n] =/= 0, so we force the rest to zero
            Y[sim, ...] = np.multiply(Yextended,np.matmul(deltaI[sim, ...], np.linalg.inv(deltaV[sim, ...])))
        else:
            Y[sim, ...] = np.matmul(deltaI[sim, ...], np.linalg.inv(deltaV[sim, ...]))

    if results_folder is not None:
        filename = results_name+'#Y_'+network.scan_type+"#"+"#".join([zblocks[idx].name+"-"+sides[idx] for idx in range(len(sides))])
        results = [Y[:, row, col] for row in range(N) for col in range(N)]
        results.insert(0, frequencies)
        elements = [str(row)+"-"+str(col) for row in range(N) for col in range(N)]
        results = tuple(results)
        header = "f\t"+"\t".join(network.all_scans)
        np.savetxt(results_folder+r'\\'+filename+'#.txt',np.stack(results, axis=1),  delimiter='\t', header=header,comments='') #, comments="\t".join(network.all_scans)

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
        for row in range(N):
            for col in range(N):
                if network.enforce:
                    # Only plot the non-zero elements
                    if Yextended[row,col] == 1:
                        ax[0].scatter(frequencies, 20*np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                else:
                    ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y[:,row,col])),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
        # ax[0].set_yscale("log")
        ax[0].set_xscale("log")
        ax[0].set_xlim([frequencies[0], frequencies[-1]])
        ax[0].minorticks_on()
        ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[0].set_ylabel('Magnitude [dB]')
        ax[0].set_title('DUT admittance ― ' + str(len(frequencies)) + ' scanned frequencies')
        ax[0].legend(loc='upper right', ncol=4)
        # ,fancybox=True, shadow=True,
        for row in range(N):
            for col in range(N):
                if network.enforce:
                    if Yextended[row,col] == 1:
                        ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
                else:
                    ax[1].scatter(frequencies,np.angle(Y[:,row,col],deg=True),linewidths=1.0,label=r'$Y_{'+str(row)+str(col)+'}$')
        ax[1].set_xscale("log")
        ax[1].set_ylim([-200, 200])
        ax[1].set_yticks([-180, -90, 0, 90, 180])
        ax[1].set_xlim([frequencies[0], frequencies[-1]])
        ax[1].minorticks_on()
        ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[1].set_ylabel('Phase [°]')
        ax[1].set_xlabel('Frequency [Hz]')
        fig.savefig(results_folder+'\\'+filename + ".pdf", format="pdf", bbox_inches="tight")
        fig.clear()


def SISO_TF(f_base=None, frequencies=None, fft_periods=1, dt=None,
            start_idx=None, zblocks=None, results_folder=None, results_name='SISO_TF'):
    # Small-signal sinusoidal steady state computation and rFFT (no target frequency-based FFT distinction)
    L = int(fft_periods * 1 / f_base * 1.0 / dt)  # For the FFT computation

    # For each simulation (freq) a transfer function is computed
    delta_input = np.empty(len(frequencies), dtype='cdouble')  # Also dtype='csingle'
    delta_output= np.empty(len(frequencies), dtype='cdouble')
    Y = np.empty(len(frequencies), dtype='cdouble')

    # Build the small-signal complex voltage and current matrices at every frequency
    for sim, frequency in enumerate(frequencies):
        fft_idx = int(round(frequency * fft_periods * 1 / f_base))  # Index of the target FFT frequency
        for block in zblocks:
            for name in ['inputTF', 'outputTF']:
                delta = block.perturbation_data[sim][name+"_TF"][start_idx:] - block.snapshot_data[name][start_idx:]
                delta_FD = np.fft.rfft(delta, n=L, axis=0) * 2 / L
                # Retrieve the response at the target frequency
                if "input" in name:
                    delta_input[sim] = delta_FD[fft_idx]
                else:
                    delta_output[sim] = delta_FD[fft_idx]

            Y[sim] =delta_output[sim] / delta_input[sim]

    if results_folder is not None:
        filename = results_name+'#SISO_TF_'+"#"+zblocks[0].name
        results = [Y]
        results.insert(0, frequencies)
        results = tuple(results)
        header = "f\t"+"\t"+zblocks[0].name
        np.savetxt(results_folder+r'\\'+filename+'#.txt',np.stack(results, axis=1),  delimiter='\t', header=header,comments='')

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
        ax[0].scatter(frequencies, 20 * np.log10(np.abs(Y)),linewidths=1.0,label=str(zblocks[0].name))
        ax[0].set_xscale("log")
        ax[0].set_xlim([frequencies[0], frequencies[-1]])
        ax[0].minorticks_on()
        ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[0].set_ylabel('Magnitude [dB]')
        ax[0].set_title('SISO Transfer function ― ' + str(len(frequencies)) + ' scanned frequencies')
        ax[0].legend(loc='upper right', ncol=4)
        
        ax[1].scatter(frequencies,np.angle(Y,deg=True),linewidths=1.0,label=str(zblocks[0].name))
        ax[1].set_xscale("log")
        ax[1].set_ylim([-200, 200])
        ax[1].set_yticks([-180, -90, 0, 90, 180])
        ax[1].set_xlim([frequencies[0], frequencies[-1]])
        ax[1].minorticks_on()
        ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
        ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
        ax[1].set_ylabel('Phase [°]')
        ax[1].set_xlabel('Frequency [Hz]')
        fig.savefig(results_folder+'\\'+filename + ".pdf", format="pdf", bbox_inches="tight")
        fig.clear()


admittance.__doc__ = """
Compute the admittance at every frequency based on the perturbation data obtained with the "frequency_sweep" function.

The function iterates over the time-domain simulation results for every frequency stored in "zblocks" and computes the real-side FFT of the perturbed
waveforms under sinusoidal steady-state. The small-signal voltages and currents at every frequency matrices are built and the admittance is simply
computed as Currents * inv(Voltages). Additional options allow the user to compute the dq-frame admittance with half of the perturbations needed
for a general NxN matrix at each frequency by assuming dq-frame symmetry. 
The function accepts several arguments:

Required
    f_base          Base frequency (determines frequency resolution) [Hz]
    frequencies     List of frequencies at which the perturbations are performed [Hz]
    scantype        To distinguish between the following subsystems
                        single AC bus: scantype = "AC" (default), e.g. three-phase system bus
                        single DC node: scantype = "DC", e.g. DC system node
                        single AC bus and single DC node: cantype = "ACDC", e.g. AC/DC converter
                        generic AC, DC and/or AC/DC multi-terminal: scantype = "Network"
    zblocks         Instance of the Scanblock class in frequency_sweep.py, or list thereof, storing information from the PSCAD Z-tool library blocks such as EMT simulation data.
    sides           Sides of the PSCAD Z-tool library blocks at which the admittance is to be computed.
                    sides is a single number for single AC bus or single DC node cases, while it is a list otherwise.
                    The list contains the side of each block as apearing in the same order as in zblocks.
    dt              Simulation sampling time used to compute the FFT [s]
    start_idx       Index of the time-domain waveforms corresponding to the sinusoidal steady-state after the injections.
    results_folder  Absolute path where the admittance is to be stored.

Optional
    fft_periods     Number of periods used to compute the FFT. Default = 1.
    results_name    Default = "Y".
    network         Instance of the Network class in frequency_sweep.py containing information related to the type of subnetwork, blocks involved and their sides,
                    topology or connectivity, perturbations performed at each EMT run, block identifiers, etc. 
    exploit_dq_sym  The dq-frame symmetry of AC subsystems is exploited when set to True so as to compute their admittance matrix with half of the number of simulations. Default = False.
    
"""


"""
This file contains a couple of functions to read and save the output files from PSCAD
Warning: Documentation for this function is not updated
"""

"""
Copyright (C) 2024  Francisco Javier Cifuentes Garcia

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__all__ = ['single_s', 'multiple_s']

import numpy as np  # Numerical python functions
from os import listdir

AC_scan_variables = ['IDUTacA1:1', 'IDUTacA1:2','VDUTac:1', 'VDUTac:2', 'IDUTacA2:1', 'IDUTacA2:2']
DC_scan_variables = ['IDUTdcA1', 'IDUTdcA2','VDUTdc']
scan_variables = AC_scan_variables + DC_scan_variables

def single_s(out_files=None, save_folder=None, save=False, files=None, zblocks=None, new_file_name=None, scan_vars=scan_variables):
    for file_num in files:
        # Load each target file and for each block related to the file asign the corresponding data to the block
        if file_num < 10:  # If the file number is less than 10, then it adds 0 before the file number
            values = np.loadtxt(out_files + "_0" + str(file_num) + ".out", skiprows=1)
        else:
            values = np.loadtxt(out_files + "_" + str(file_num) + ".out", skiprows=1)
        for block in zblocks:
            block.snapshot_data["time"] = values[:, 0]  # Retreive the time vector
            for block_file in block.files_to_open:
                if file_num == block_file:
                    # If the block needs data from the file, then use the block's target columns for this file
                    for col in block.relative_cols[file_num]:
                        ch = col - 1 + 10 * (file_num - 1)  # Absolute output channel number
                        # print(block.out_vars_names[ch], ch_var_names[ch])
                        block.snapshot_data[block.out_vars_names[ch]] = values[:, col - 1]  # Retreived data

    if save:
        new_file_name = save_folder + '\\' + new_file_name + '.txt'
        var_names = ["time"]
        data = zblocks[0].snapshot_data["time"].reshape(-1, 1)  # Retreive the time vector data
        for block in zblocks:
            for name in list(block.out_vars_names.values()):
                if (name != "time") and (name in scan_vars):  # Only currents and voltages
                    data = np.append(data, block.snapshot_data[name].reshape(-1, 1), axis=1)
                    var_names.append(name)
        np.savetxt(new_file_name, data, delimiter='\t', header="\t".join(var_names))

def multiple_s(n_sim=None, out_folder=None, file_name=None, save_folder=None, save=False, tar_files=None, zblocks=None, scan_vars=scan_variables):
    # # Filter the file names to identify the target multiple output files
    # files = [file for file in listdir(out_folder) if
    #          (file.endswith(".out") and (file.count(file_name) > 0))]  # end in .out and contain name
    # # More file filtering: file name followed by an _ and another two _ near the end
    # files_filtered = [file for file in files if (file[len(file_name)] == '_' and
    #                                              file.count("_", len(file_name)) == 2 and
    #                                              len(file) > len(file_name))]
    # # Sort the files from low to high simulation: split by '_' and take the rank number (position 2 from the end)
    # files_filtered.sort(key=lambda file_name: int(file_name.split('_')[-2]))
    # for i in files_filtered: print(i)  # Just for debugging

    # New function
    save_time = True  # Save the time vector only once improving the memory usage
    root_name = out_folder + '\\' + file_name  # Root of the filename
    sim_type = "_"+file_name.split("_")[-1]  # Use the end of the file name (sim type) to re-name the output variables
    # print("Perturbation type: ",sim_type)
    for sim in range(1,n_sim+1):
        # For each simulation
        for file_num in tar_files:
            # Load each target file and for each block related to the file and asign the corresponding data to the block
            if file_num < 10:  # If the file number is less than 10, then it adds 0 before the file number
                if sim < 10:  # If the simulation number is less than 10, then it adds a 0 before the simultation number
                    values = np.loadtxt(root_name + "_0" + str(sim) + "_0" + str(file_num) + ".out", skiprows=1)
                else:
                    values = np.loadtxt(root_name + "_" + str(sim) + "_0" + str(file_num) + ".out", skiprows=1)
            else:
                if sim < 10:
                    values = np.loadtxt(root_name + "_0" + str(sim) + "_" + str(file_num) + ".out", skiprows=1)
                else:
                    values = np.loadtxt(root_name + "_" + str(sim) + "_" + str(file_num) + ".out", skiprows=1)
            # if save_time:
            #     zblocks[0].perturbation_data["time"] = values[:, 0]  # Retreive the time vector
            #     # print(" Time vector start",zblocks[0].perturbation_data["time"][0],"and end",zblocks[0].perturbation_data["time"][-1])
            #     save_time = False  # Only for the first scanning block (to save memory)
            for block in zblocks:
                block.perturbation_data["time"] = values[:, 0]  # Retreive the time vector
                # For each z-tool block related to the file, asign the corresponding data to the block
                for block_file in block.files_to_open:
                    if file_num == block_file:
                        # print("Simulation",sim, "block:", block.name, 'File:', block_file)
                        # If the block needs data from the file, then use the block's target columns for this file
                        for col in block.relative_cols[file_num]:
                            ch = col - 1 + 10 * (file_num - 1)  # Absolute output channel number for the column
                            # print(" Channel",ch,"variable",block.out_vars_names[ch],"initial value",values[0, col - 1])
                            if block.out_vars_names[ch] in scan_vars:
                                # Retrieve the currents and voltages
                                block.perturbation_data[sim-1][block.out_vars_names[ch]+sim_type] = values[:, col - 1]

    if save:
        file_name = save_folder+'\\'+file_name
        for sim in range(n_sim):
            var_names = ["time"]
            data = zblocks[0].perturbation_data["time"].reshape(-1, 1)  # Retreive the time vector data
            for block in zblocks:
                for name in list(block.out_vars_names.values()):
                    if (name != "time") and (name in scan_vars):  # Only currents and voltages
                        data = np.append(data, block.perturbation_data[sim][name+sim_type].reshape(-1, 1), axis=1)
                        var_names.append(name)
            np.savetxt(file_name + "_" + str(sim) + '.txt', data, delimiter='\t', header="\t".join(var_names))


multiple_s.__doc__ = """
Function that reads Multiple Simulations results and saves them into a dedicated file and/or loads them into memory for further processing.
The function accepts several input arguments to customize the reading:
Required
        n_sim		 Number of simulation results to be read: total number of simulations.
        original_folder	 Absolute path of the folder where the results can be found.
        file_name	 Common name of the files to be read.       

Optional
        output		 Boolean variable to control if the read results should be returned as an array. output = True returns an array with the results.
        save		 Boolean variable to control the saving of the read data. False means the data is not saved.
        new_folder	 Absolute path of the destination folder in case save = True.
        output_filename	 Output file name: no extension needed. By default it saves the data into .txt files where the first column is the time and the rest
                         are the concatenated extracted variables from each simulation in the order that they where performed.

Furthermore, there is an additional internal parameter that indicates how many variables are read from each file. For impedance computation
only two currents and voltages are needed, thus n = 4 variables.

"""
single_s.__doc__ = """
Function that reads Single Simulation results and saves them into a dedicated file and/or loads them into memory for further processing.
The function accepts several input arguments to customize the reading:
Required
        original_folder	 Absolute path of the folder where the results can be found.
        file_name	 Common name of the file to be read.       

Optional
        output		 Boolean variable to control if the read results should be returned as an array. output = True returns an array with the results.
        save		 Boolean variable to control the saving of the read data. False means the data is not saved.
        new_folder	 Absolute path of the destination folder in case save = True.
        output_filename	 Output file name: no extension needed. By default it saves the data into .txt files where the first column is the time and the rest
                         are the extracted variables.

Furthermore, there is an additional internal parameter that indicates how many variables are read from each file. For impedance computation
only two currents and voltages are needed, thus n = 4 variables.

"""



"""
Functions to convert the addmittance matrices from one frame to another
"""
"""
Copyright (C) 2024  Francisco Javier Cifuentes Garcia

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['dq_lag2dq_lead','dcdq_lag2dcdq_lead','dq2MSD','dcdq2MSD','ab2pn']

import numpy as np  # Numerical python functions

def dq_lag2dq_lead(Y_old_frame=None, frequencies=None, results_folder=None, file_name=None):
    # Transformation from dq-frame matrices with q-axis lagging to leading
    # The rotation matrix is the product of the abc-to-dq0 transform with q-axis lagging (T1) times the
    # dq0-to-abc transform with q-axis leading (T2, i.e. inverse of the abc-to-dq0 transform with q-axis leading)
    # and then negglecting the zero-sequence component (last row and column): R = T1 @ T2
    R = np.array([[1, 0], [0, -1]]) 
    Y_new_frame = np.matmul(R,np.matmul(Y_old_frame, R))  # = R @ Y @ R is equivalent to multiplying the off-diagonals of Y by -1
    # Save the new matrix
    if file_name is not None and results_folder is not None and frequencies is not None:
        np.savetxt(results_folder+'\\'+file_name+'_dq_leading.txt', np.c_[frequencies,Y_new_frame.reshape(Y_new_frame.shape[0], -1)],
                   delimiter='\t', header="f\t"+"\t".join(["dd","dq","qd","qq"]), comments='')
    return Y_new_frame

def dq2MSD(Y_old_frame=None, frequencies=None, results_folder=None, file_name=None, q_lagging=True):
    # Transformation from a dq-frame matrix to the Modified Sequence Domain (MSD) as described in
    # A. Rygg, M. Molinas, C. Zhang and X. Cai, "A Modified Sequence-Domain Impedance Definition and Its Equivalence to
    # the dq-Domain Impedance Definition for the Stability Analysis of AC Power Electronic Systems," in IEEE Journal of
    # Emerging and Selected Topics in Power Electronics, vol. 4, no. 4, pp. 1383-1396, Dec. 2016, doi: 10.1109/JESTPE.2016.2588733.

    # If the given matrix has q-axis lagging then it is firstly rotated to q-axis leading to match the reference formula
    if q_lagging:
        R = np.array([[1, 0], [0, -1]])
    else:
        R = np.eye(2)
    A = 0.5 * np.array([[1, 1j], [1, -1j]]) @ R
    A_inv = R @ np.array([[1, 1], [-1j, 1j]])
    Y_new_frame = np.matmul(A,np.matmul(Y_old_frame, A_inv))  # Y' = A @ Y @ A^(-1)
    # Save the new matrix
    if file_name is not None and results_folder is not None and frequencies is not None:
        np.savetxt(results_folder+'\\'+file_name+'_MSD_from_dq.txt', np.c_[frequencies,Y_new_frame.reshape(Y_new_frame.shape[0], -1)],
                   delimiter='\t', header="f\t"+"\t".join(["pp","pn","np","nn"]), comments='')
    return Y_new_frame

def dcdq_lag2dcdq_lead(Y_old_frame=None, frequencies=None, results_folder=None, file_name=None):
    # Transformation from dc,d,q-frame matrices with q-axis lagging to q-axis leading
    # The rotation matrix is the product of the abc-to-dq0 transform with q-axis lagging (T1) times the
    # dq0-to-abc transform with q-axis leading (T2, i.e. inverse of the abc-to-dq0 transform with q-axis leading)
    # then negglecting the zero-sequence component (last row and column), R = T1 @ T2, and considering dc-variables unaffected
    R = np.array([[1, 0, 0],[0, 1, 0], [0, 0, -1]]) 
    Y_new_frame = np.matmul(R,np.matmul(Y_old_frame, R))  # = R @ Y @ R is equivalent to multiplying the q-axis off-diagonals of Y by -1
    # Save the new matrix
    if file_name is not None and results_folder is not None and frequencies is not None:
        np.savetxt(results_folder+'\\'+file_name+'_dq_leading.txt', np.c_[frequencies,Y_new_frame.reshape(Y_new_frame.shape[0], -1)],
                   delimiter='\t', header ="f\tdc-dc\tdc-d\tdc-q\td-dc\td-d\td-q\tq-dc\tq-d\tq-q", comments='')
    return Y_new_frame

def dcdq2MSD(Y_old_frame=None, frequencies=None, results_folder=None, file_name=None, q_lagging=True):
    # Transformation from dc,d,q-frame admittance to the Modified Sequence Domain (MSD) as described in
    # S. Shah and L. Parsa, "Sequence domain transfer matrix model of three-phase voltage source converters," 2016 IEEE
    # Power and Energy Society General Meeting (PESGM), Boston, MA, USA, 2016, pp. 1-5, doi: 10.1109/PESGM.2016.7742009.

    # If the given matrix has q-axis lagging then it is firstly rotated to q-axis leading to match the reference formula
    if q_lagging:
        R = np.array([[1, 0], [0, -1]])
    else:
        R = np.eye(2)
    A = 0.5 * np.array([[1, 1j], [1, -1j]]) @ R
    A_inv = R @ np.array([[1, 1], [-1j, 1j]])
    A = np.block([[1,np.zeros((1,2))],[np.zeros((2,1)), A]])
    A_inv = np.block([[1,np.zeros((1,2))],[np.zeros((2,1)), A_inv]])
    Y_new_frame = np.matmul(A,np.matmul(Y_old_frame, A_inv))  # Y' = A @ Y @ A^(-1)
    # Save the new matrix
    if file_name is not None and results_folder is not None and frequencies is not None:
        np.savetxt(results_folder+'\\'+file_name+'_MSD_from_3x3.txt', np.c_[frequencies,Y_new_frame.reshape(Y_new_frame.shape[0], -1)],
                   delimiter='\t', header="f\t"+"\t".join(["dc,dc","dc,p","dc,n","p,dc","p,p","p,n","n,dc","n,p","n,n"]), comments='')
    return Y_new_frame

def ab2pn(Y_old_frame=None, frequencies=None, results_folder=None, file_name=None):
    A = 0.5 * np.array([[1, 1j], [1, -1j]])  # Same rotation matrix as dq2MSD()
    A_inv = R @ np.array([[1, 1], [-1j, 1j]])
    Y_new_frame = np.matmul(A, np.matmul(Y_old_frame, A_inv))  # Y' = A @ Y @ A^(-1)
    # Save the new matrix
    if file_name is not None and results_folder is not None and frequencies is not None:
        np.savetxt(results_folder + '\\' + file_name + '_pn_from_ab.txt',np.c_[frequencies, Y_new_frame.reshape(Y_new_frame.shape[0], -1)],
                   delimiter='\t', header="f\t" + "\t".join(["pp", "pn", "np", "nn"]), comments='')
    return Y_new_frame



dq_lag2dq_lead.__doc__ = """
 Transformation from dq-frame matrices with q-axis lagging to leading
 The rotation matrix is the product of the abc-to-dq0 transform with q-axis lagging (T1) times
 the dq0-to-abc transform with q-axis leading (T2, i.e. inverse of the abc-to-dq0 transform with
 q-axis leading) and then negglecting the zero-sequence component (last row and column): R = T1 @ T2

Required arguments
        Y_old_frame     Numpy matrix of Nx2x2 to be transformed where N is the number of frequency points

Optional arguments
        frequencies     Numpy vector of frequency points. Only needed if the result is to be saved.
        results_folder  Full path of the directory where the result is to be saved.
        file_name       Root name of the saved text file. By default the code adds "dq_leading" to indicate that the matrix has been transformed.
"""

dcdq_lag2dcdq_lead.__doc__ = """
 Transformation from dc,d,q-frame matrices with q-axis lagging to q-axis leading
 The rotation matrix is the product of the abc-to-dq0 transform with q-axis lagging (T1) times
 the dq0-to-abc transform with q-axis leading (T2, i.e. inverse of the abc-to-dq0 transform with
 q-axis leading), then negglecting the zero-sequence component (last row and column): R = T1 @ T2
 while the dc-variables are not afected.

Required arguments
        Y_old_frame     Numpy matrix of Nx3x3 to be transformed where N is the number of frequency points

Optional arguments
        frequencies     Numpy vector of frequency points. Only needed if the result is to be saved.
        results_folder  Full path of the directory where the result is to be saved.
        file_name       Root name of the saved text file. By default the code adds "dq_leading" to indicate that the matrix has been transformed.
"""

dq2MSD.__doc__ = """
Transformation from dq-frame matrix to the Modified Sequence Domain (MSD) as described in
A. Rygg, M. Molinas, C. Zhang and X. Cai, "A Modified Sequence-Domain Impedance Definition and Its Equivalence to
the dq-Domain Impedance Definition for the Stability Analysis of AC Power Electronic Systems," in IEEE Journal of
Emerging and Selected Topics in Power Electronics, vol. 4, no. 4, pp. 1383-1396, Dec. 2016, doi: 10.1109/JESTPE.2016.2588733.
If the given matrix has q-axis lagging then it is firstly rotated to q-axis leading to match the referenced formula

Required arguments
        Y_old_frame     Numpy matrix of Nx2x2 to be transformed where N is the number of frequency points

Optional arguments
        frequencies     Numpy vector of frequency points. Only needed if the result is to be saved.
        results_folder  Full path of the directory where the result is to be saved.
        file_name       Root name of the saved text file. By default the code adds "_MSD_from_dq" to indicate that the matrix has been transformed.
"""

dcdq2MSD.__doc__ = """
Transformation from dc,d,q-frame admittance to the Modified Sequence Domain (MSD) as described in
S. Shah and L. Parsa, "Sequence domain transfer matrix model of three-phase voltage source converters," 2016 IEEE
Power and Energy Society General Meeting (PESGM), Boston, MA, USA, 2016, pp. 1-5, doi: 10.1109/PESGM.2016.7742009.
If the given matrix has q-axis lagging then it is firstly rotated to q-axis leading to match the reference formula

Required arguments
        Y_old_frame     Numpy matrix of Nx2x2 to be transformed where N is the number of frequency points

Optional arguments
        frequencies     Numpy vector of frequency points. Only needed if the result is to be saved.
        results_folder  Full path of the directory where the result is to be saved.
        file_name       Root name of the saved text file. By default the code adds "dq_leading" to indicate that the matrix has been transformed.
"""


"""
This program contains several functions that can be used for the stability analysis in the frequency-domain, including:
    1) Basic Generalized Nyquist Criterion (GNC) application to determine system stability margins
    2) Eigenvalue Decomposition (EVD) over the frequency to determine oscillatory modes and participation factors
    3) Passivity and singular values of target matrices
    4) Sensitivity of the state-space oscillatory mode w.r.t. admittance elements (ongoing)
    5) A main stability_analysis function to apply all the previously described to a specific system

Copyright (C) 2024  Francisco Javier Cifuentes Garcia

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['stability_analysis','passivity','nyquist','small_gain','EVD','nyquist_det']

import numpy as np
from scipy.optimize import linear_sum_assignment
import matplotlib.pyplot as plt
from .read_admittance import read_admittance
from os import path, makedirs
import pickle

from matplotlib import rcParams  # Text's parameters for plots
rcParams['mathtext.fontset'] = 'cm'  # Font selection
rcParams['font.family'] = 'STIXGeneral'  # 'cmu serif'

class Graph:
    def __init__(self, V):
        self.V = V
        self.adj = [[] for i in range(V)]

    # Depth-first search algorithm method
    def DFSUtil(self, temp, v, visited):
        # Mark the current vertex as visited
        visited[v] = True
        # Add the vertex to list
        temp.append(v)
        # Repeat for all vertices adjacent to this vertex v
        for i in self.adj[v]:
            if not visited[i]:
                # Update the list
                temp = self.DFSUtil(temp, i, visited)
        return temp

    # Add an undirected edge
    def addEdge(self, v, w):
        self.adj[v].append(w)
        self.adj[w].append(v)

    # Method to retrieve connected components in an undirected graph
    def connectedComponents(self):
        visited = []
        cc = []
        for i in range(self.V):
            visited.append(False)
        for v in range(self.V):
            if not visited[v]:
                temp = []
                cc.append(self.DFSUtil(temp, v, visited))
        return cc

def stability_analysis(topology=None, results_folder=None, file_root=None, check_conditioning=False, condition_number_th=10e6):
    # 0) Read the terminal angle information
    block_area_angle = []  # List for each block containing a list as [bus/block name, area_id, terminal angle in rad]
    with open(results_folder+r'\\'+file_root+'_angles.txt', 'r') as file:
        next(file)  # First line contains the header
        for line in file:
            content = [str(line.strip().split("\t")[0]),int(line.strip().split("\t")[1]),float(line.strip().split("\t")[2])]
            block_area_angle.append(content)
    number_areas = len(set([bus[1] for bus in block_area_angle]))
    # Define a reference bus for each area
    reference_angle = {i: -100.0 for i in range(number_areas)}  # Unrealistic intial values
    for bus in block_area_angle:
        if number_areas == 1:
            reference_angle[0] = float(bus[2])
            reference_angle[1] = float(bus[2])
        else:
            if reference_angle[bus[1]] == -100.0:
                reference_angle[bus[1]] = float(bus[2])  # If no angle is already defined, use the first one from that area
                # print("The reference for area",bus[1],'is bus',bus[0])

    # This function loads and builds the edge and nodal admittance matrices and for stability analysis
    # 1) Read the topology matrix and extract block names
    Ytopology = np.loadtxt(topology, skiprows=1, comments=["#", "%", "!"])
    # nameA-1 nameA-2 nameB-1 nameB-2 ... x nameA-1 nameA-2 nameB-1 ...
    # 0 means no interconnection, 1 means connection between the edges: diagonals are single-sided / shunt
    with open(topology, 'r') as f:
        block_names_Y = f.readline().strip('\n').split("\t")

    # 2) Read the admittance files based on the topology file
    admittances = []  # List containing the admittance objects
    bus_names = []  # List of list with block (bus) names
    # Create the undirected graph: adjacent matrix but diagonals can be 1
    g = Graph(len(Ytopology))
    for row, name in enumerate(block_names_Y):
        for col, edge in enumerate(Ytopology[row]):
            if int(edge) == 1: g.addEdge(row, col)
    # Obtain the connected components of the graph
    cc = g.connectedComponents()  # List of lists with blocks positions connected
    # print("Connected components",cc)
    # For every group of connected buses, read and add the admittance matrix
    for buses in cc:
        bus_names.append([block_names_Y[bus] for bus in buses])
        # print("Bus names", bus_names)
        admittances.append(read_admittance(path=results_folder, involved_blocks=bus_names[-1], file_root=file_root))

    # 3) Update bus_names to be ordered as the variables in the individual admittance matrices and build the node matrix
    node_matrix = []  # Create the node matrix with the active components (block diagonal)
    node_variables = []  # List of variable names = the current/voltage vectors
    edge_aux_matrix = []  # Auxiliary edge matrix (block diagonal)
    edge_aux_variables = []  # The order of this aux matrix is different from the order of the node matrix
    for idx, y in enumerate(admittances):
        bus_names[idx] = y.blocks
        # print(bus_names[idx],y.y_type,"- Node:",y.node)
        if y.node:
            # Define the nodal matrix that sets the order of the electrical variables
            for var in y.vars: node_variables.append(var)
            node_matrix.append(y)
        else:
            # Update the aux edge matrix and its variables
            for var in y.vars: edge_aux_variables.append(var)
            edge_aux_matrix.append(y)

    print("Node vars \n",node_variables,"\nEdge vars \n",edge_aux_variables)
    edge_ordering = []  # List to re-sort the edge matrix acording to the node matrix variables
    for var in node_variables: edge_ordering.append(edge_aux_variables.index(var))
    # print("Node vars \n",node_variables,"\nEdge vars \n",edge_aux_variables)
    # print("\nSorted edge variables \n",sorted(edge_aux_variables,key=node_variables.index))

    # 4) Create the node and edge matrices with the frequency domain data
    frequencies = admittances[0].f  # Retreive the frequency vector
    # Create the auxiliary edge matrix (different order than node matrix), useful to check the network topology and scan
    Yedge_aux = np.zeros((len(frequencies),len(node_variables),len(node_variables)),dtype='cdouble')  # Or dtype='csingle'
    y_edge_idx = 0
    for idx, yedge in enumerate(edge_aux_matrix):
        # # Eliminate too small elements: related to PSCAD accuracy when the topology is not enforced
        # for col in range(np.size(yedge.y, 1)):
        #     for row in range(np.size(yedge.y, 2)):
        #         if max(abs(yedge.y[:, row, col])) < 1e-6:  # The threshold is system and time-step dependent
        #             yedge.y[:, row, col] = np.zeros(yedge.y[:, row, col].shape)
        Yedge_aux[:, y_edge_idx:y_edge_idx + len(yedge.vars), y_edge_idx:y_edge_idx + len(yedge.vars)] = yedge.y
        y_edge_idx = y_edge_idx + len(yedge.vars)  # Update the matrix index for the next admittance block

    # Sparsity plot for verification at the lowest frequency
    np.seterr(divide='ignore')
    plt.imshow(20 * np.log10(np.abs(Yedge_aux[0, :, :])), cmap='spring', interpolation='nearest')
    plt.colorbar()
    plt.xticks(ticks=np.arange(0, len(edge_aux_variables), step=1))
    plt.yticks(ticks=np.arange(0, len(edge_aux_variables), step=1),labels=edge_aux_variables)
    plt.grid(visible=True, which='minor', alpha=0.3, color='k', linestyle='-', linewidth=0.5)
    plt.title('Auxiliary edge admittance matrix at '+format(frequencies[0], '.1f')+' Hz')
    plt.savefig(results_folder + '\\' + file_root + "_Edge_aux.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + file_root + "_Edge_aux.pickle", 'wb') as f:
        pickle.dump(plt.gcf(), f)
    plt.close()

    # Re-sort the edge matrix acording to the node matrix variables
    Yedge = Yedge_aux[:,:,edge_ordering]  # Sort the columns
    Yedge = Yedge[:,edge_ordering,:]  # Sort the rows

    # Plot the sorted edge admittance for verification at the lowest frequency
    plt.imshow(20 * np.log10(np.abs(Yedge[0, :, :])), cmap='spring', interpolation='nearest')
    plt.colorbar()
    plt.xticks(ticks=np.arange(0, len(node_variables), step=1))
    plt.yticks(ticks=np.arange(0, len(node_variables), step=1), labels=node_variables)
    #  plt.minorticks_on()
    plt.grid(visible=True, which='minor', alpha=0.3, color='k', linestyle='-', linewidth=0.5)
    plt.title('Edge admittance matrix at '+format(frequencies[0], '.1f')+' Hz')
    plt.savefig(results_folder + '\\' + file_root + "_Edge.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + file_root + "_Edge.pickle", 'wb') as f:
        pickle.dump(plt.gcf(), f)
    plt.close()

    # Node admittance matrix and associated edge matrix
    Ynode = np.zeros((len(frequencies), len(node_variables), len(node_variables)),dtype='cdouble')  # Or dtype='csingle'
    y_node_idx = 0
    for idx, ynode in enumerate(node_matrix):
        # Rotate the matrix accordingly to the terminal angle if it involves AC variables
        if len(ynode.vars) >= 2:
            for block in block_area_angle:
                # Iterate over the different blocks and use the angle of the current matrix
                if block[0] in [bus[:-2] for bus in ynode.blocks]:
                    # The reference angle for the area of this block - bus angle of this block
                    theta = (reference_angle[block[1]] - block[2])
                    # print("Angle of",round(theta,4),"in the admittance involving",ynode.blocks)

        # Define the rotation matrix for each component: AC-side 2x2, AC/DC 3x3
        if len(ynode.vars) == 2:
            T = np.array([[np.cos(theta), -np.sin(theta)],[np.sin(theta), np.cos(theta)]])
            T_1 = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
        elif len(ynode.vars) == 3:
            T = np.array([[1, 0, 0],[0, np.cos(theta), -np.sin(theta)],[0, np.sin(theta), np.cos(theta)]])
            T_1 = np.array([[1, 0, 0], [0, np.cos(theta), np.sin(theta)], [0, -np.sin(theta), np.cos(theta)]])
        else:
            T = np.identity(len(ynode.vas))  # DC-side analysis only: does not need frame alignment
            T_1 = np.identity(len(ynode.vars))

        # print('Matrix',idx,", nodal matrix start idx:",y_node_idx,"and end:",y_node_idx+len(ynode.vars))
        Ynode[:, y_node_idx:y_node_idx+len(ynode.vars), y_node_idx:y_node_idx+len(ynode.vars)] = T@ynode.y@T_1
        # Equivalent to np.matmul(T,np.matmul(ynode.y,T_1))
        y_node_idx = y_node_idx + len(ynode.vars)  # Update the matrix index for the next admittance block

    # Scatter verification of the node admittance for the lowest frequency
    plt.imshow(20*np.log10(np.abs(Ynode[0, :, :])), cmap='spring', interpolation='nearest')
    plt.colorbar()
    plt.xticks(ticks=np.arange(0, len(node_variables), step=1))
    plt.yticks(ticks=np.arange(0, len(node_variables), step=1),labels=node_variables)
    #  plt.minorticks_on()
    plt.grid(visible=True, which='minor', alpha=0.3, color='k', linestyle='-', linewidth=0.5)
    plt.title('Node admittance matrix at '+format(frequencies[0], '.1f')+' Hz')
    plt.savefig(results_folder + '\\' + file_root + "_Node.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + file_root + "_Node.pickle", 'wb') as f:
        pickle.dump(plt.gcf(), f)
    plt.close()
    np.seterr(divide='warn')

    # Perform stability analysis
    L = np.matmul(np.linalg.inv(Yedge),Ynode)  # Loop gain matrix
    nyquist(L,frequencies,results_folder,file_root,check_conditioning,condition_number_th)  # Stability via eigenvalue loci
    nyquist_det(L,frequencies,results_folder,file_root, verbose=False, offset=0.0, draw_arrows=True, show_plot=False)  # Stability via determinant

    # Oscillatory frequencies and participation factors based on the closed-loop impedance matrix (bus)
    EVD(np.linalg.inv(Yedge+Ynode), frequencies, node_variables, results_folder, file_root)

    # Save the admittance matrices
    results = [Yedge[:, row, col] for row in range(len(node_variables)) for col in range(len(node_variables))]
    results.insert(0, frequencies)
    results = tuple(results)
    np.savetxt(results_folder + '\\' + file_root + '_Y_edge.txt', np.stack(results, axis=1), delimiter='\t',
               header="f\t" + "\t".join(node_variables), comments='')

    results = [Ynode[:, row, col] for row in range(len(node_variables)) for col in range(len(node_variables))]
    results.insert(0, frequencies)
    results = tuple(results)
    np.savetxt(results_folder + '\\' + file_root + '_Y_node.txt', np.stack(results, axis=1), delimiter='\t',
               header="f\t" + "\t".join(node_variables), comments='')

    # Save the minor loop gain matrix
    results = [L[:, row, col] for row in range(len(node_variables)) for col in range(len(node_variables))]
    results.insert(0, frequencies)
    # elements = [str(row) + "-" + str(col) for row in range(len(node_variables)) for col in range(len(node_variables))]
    results = tuple(results)
    np.savetxt(results_folder+'\\'+file_root+'_Minor_loop_gain.txt',np.stack(results, axis=1),delimiter='\t',
               header="f\t" + "\t".join(node_variables), comments='')

    # Compute the passivity and SVD of the system matrices
    # passivity(G=Yedge_aux, frequencies=frequencies, results_folder=results_folder, filename=file_root + "_Yedge_aux",variables=edge_aux_variables)
    passivity(G=Ynode,frequencies=frequencies,results_folder=results_folder,filename=file_root+"_Ynode",variables=node_variables)
    passivity(G=Yedge, frequencies=frequencies, results_folder=results_folder, filename=file_root + "_Yedge")
    passivity(G=Ynode+Yedge, frequencies=frequencies, results_folder=results_folder, filename=file_root+"_Ynode_+_Yedge")
    small_gain(np.linalg.inv(Yedge), Ynode, frequencies, results_folder, file_root,variables=node_variables)

def passivity(G, frequencies, results_folder=None, filename=None, variables=None, Yedge=None):
    # The passivity index is computed as half of the minimum eigenvalue of the matrix plus its conjugate transpose
    # min{eig(A + A')}/2
    # A passive system has its Nyquist plot in the RHP: increased stability margins if connected to a passive system
    passivity_index = np.real(np.min(np.linalg.eig(G + G.swapaxes(-1, -2).conj())[0], axis=1))/2  # min{eig(A + A')}/2
    # The eigenvalues of a Hermitian matrix are always real but floating-point arithmetic renders a small complex part
    # and thus we take the real value at the end to get rid of the spurious numerical artifact.

    if (results_folder is not None) and (filename is not None):
        if not path.exists(results_folder): makedirs(results_folder)  # Create results folder if it does not exist
        
        # Plot the passivity index over the frequency range
        if variables is None:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            ax.plot(frequencies, passivity_index, color='blue', linestyle='solid', linewidth=2.0,label=r"$\mathbf{Y}_{node}$")
            ax.set_xscale("log")
            ax.minorticks_on()
            ax.grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax.grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax.set_ylabel(r'min $\{ \lambda (\mathbf{G} + \mathbf{G}^H) \}/2$')
            ax.set_title('Passivity evaluation for ' + str(len(frequencies)) + ' frequencies')
            ax.set_xlim([frequencies[0], frequencies[-1]])
            ax.set_xlabel('Frequency [Hz]')
            if Yedge is not None:
                passivity_index_Yedge = np.real(np.min(np.linalg.eig(Yedge+Yedge.swapaxes(-1, -2).conj())[0], axis=1))/2
                ax.plot(frequencies, passivity_index_Yedge, color='red', linestyle='solid', linewidth=2.0,label=r"$\mathbf{Y}_{edge}$")
                ax.plot(frequencies, passivity_index + passivity_index_Yedge, color='green', linestyle='dashed', linewidth=2.0,label=r'min $\{ \lambda (\mathbf{Y}_{node} + \mathbf{Y}_{node}^H) \}/2$+min $\{ \lambda (\mathbf{Y}_{edge} + \mathbf{Y}_{edge}^H) \}/2$')
                ax.plot(frequencies, np.real(np.min(np.linalg.eig(G + G.swapaxes(-1, -2).conj() + Yedge + Yedge.swapaxes(-1, -2).conj())[0],axis=1))/2, color='black', linestyle='dotted',linewidth=2.0, label=r"$\mathbf{Y}_{node}+\mathbf{Y}_{edge}$")

                ax.legend(loc='upper left', fancybox=True, shadow=True, ncol=2)
        else:
            # Find the position of the block diagonal matrices as they are surrounded by zeros
            indices = []  # Tuple of start and end indeces of each matrix
            start_index = 0
            for index in range(G.shape[1]-1):
                if G[0, index, start_index] == 0:  # If the next element is zero, then boundary of block matrix if defined
                    indices.append((start_index, index-1))  # There is a block matrix between start_index and index
                    start_index = index  # Initialize start index of next matrix
            indices.append((start_index, G.shape[1] - 1))  # Last block matrix

            # Plot the passivity of the different matrices and the whole matrix
            fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
            for block_pos in range(len(indices)):
                start_idx = indices[block_pos][0]
                end_idx = indices[block_pos][1] + 1
                A = G[:, start_idx:end_idx, start_idx:end_idx]
                passivity_index_block = np.real(np.min(np.linalg.eig(A + A.swapaxes(-1, -2).conj())[0], axis=1))/2
                ax[0].plot(frequencies, passivity_index_block, linestyle='solid', linewidth=2.0,label=", ".join(variables[start_idx:end_idx]))

            ax[0].set_xscale("log")
            ax[0].minorticks_on()
            ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel(r'min $\{ \lambda (\mathbf{G}_i + \mathbf{G}_i^H) \}/2$')
            ax[0].set_title('Passivity analysis for ' + str(len(frequencies)) + ' frequencies')
            ax[0].set_xlim([frequencies[0], frequencies[-1]])
            ax[0].legend(loc='upper left', fancybox=True, shadow=True, ncol=1)

            ax[1].plot(frequencies,passivity_index,color='blue',linestyle='solid',linewidth=2.0,label="Complete matrix")
            ax[1].set_xscale("log")
            ax[1].minorticks_on()
            ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
            ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
            ax[0].set_ylabel(r'min $\{ \lambda (\mathbf{G} + \mathbf{G}^H) \}/2$')
            ax[1].legend(loc='upper left', fancybox=True, shadow=True, ncol=1)
            ax[1].set_xlim([frequencies[0], frequencies[-1]])
            ax[1].set_xlabel('Frequency [Hz]')
        
        fig.savefig(results_folder+'\\'+filename + "_passivity.pdf", format="pdf", bbox_inches="tight")
        with open(results_folder + '\\' + filename + "_passivity.pickle", 'wb') as f:
            pickle.dump(fig, f)
        np.savetxt(results_folder+'\\'+filename+'_passivity.txt', np.stack((frequencies, passivity_index), axis=1), delimiter='\t',
                   header="f\t" + "Passivity_index", comments='')
        plt.close(fig)

    return passivity_index

def nyquist(L, frequencies, results_folder=None, filename=None, verbose=True, check_conditioning=False, condition_number_th=0.01/5e-9):
    # Stability assessment: aplication of Generalized Nyquist Criteria (GNC)
    # N: Net number of clockwise encirclements by the loop gain eigenloci = # clockwise - # counter-clockwise
    # P: Number of RHP poles of the loop gain
    # Z: number of RHP poles of the closed-loop system
    # Argument principle: N = Z - P
    # If Z = N + P > 0, then the closed-loop system is unstable
    # Assuming standalone stable subsystems means P = 0, thus N = 0 implies stability while N > 0 implies instability
    # If the subsystems are standalone unstable, P > 0, then possibly N < 0

    if not path.exists(results_folder): makedirs(results_folder)  # Create results folder if it does not exist

    # 1) Compute the eigenvalues of the loop-gain at every frequency
    eigenvalues = np.linalg.eig(L)[0]
    # Also compute the condition number to discard erroneous data: threshold based on input error
    if check_conditioning: condition_number = np.linalg.cond(L)  # Relative error out <= cond_num * relativer error input

    # Sorting of eigenvalues for a continuous eigenloci in the Nyquist plot based on minimum changes between frequencies
    eigenvalues_sorted = np.empty(eigenvalues.shape, dtype='cdouble')  # Initialization of sorted eigenvalues
    eigenvalues_sorted[0,:] = eigenvalues[0,:]  # First frequency as reference for the sorting
    for idx in range(1,eigenvalues.shape[0]):
        if check_conditioning:
            if condition_number[idx] > condition_number_th:
                eigenvalues_sorted[idx, :] = eigenvalues_sorted[idx-1, :] # Replace by previous well-conditioned values
        else:
            eig_1 = eigenvalues_sorted[idx - 1, :]  # Previous eigenvalues
            eig_2 = eigenvalues[idx, :]  # Current eigenvalues
            # Create matrix of distances between eigenvalues
            x, y = np.meshgrid(np.real(eig_1), np.real(eig_2), indexing='ij')
            d_real = np.abs(x - y)
            x, y = np.meshgrid(np.imag(eig_1), np.imag(eig_2), indexing='ij')
            d_imag = np.abs(x - y)
            d_abs = np.sqrt(np.square(d_real) + np.square(d_imag))  # Absolute distance by element-wise operations
            # Solve the linear sum assignment problem to find the minimum variation and thus the correct order
            col_ind = linear_sum_assignment(d_abs)[1]  # The absolute distance is the cost matrix
            eigenvalues_sorted[idx, :] = eigenvalues[idx, col_ind]  # Sort the eigenvalues

    # 2) Eigenloci plot and count clockwise and counter-clockwise encirclements of (-1,0j) for each eigenvalue
    # Compute the coordinates of the eigenloci for the GNC aplication and plotting
    x = np.real(eigenvalues_sorted)  # Real axis
    y = np.imag(eigenvalues_sorted)  # Imaginary axis
    cw = []  # List of clockwise encirclements for each locus
    ccw = []  # List of counter-clockwise encirclements for each locus
    fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(6, 7))  # Create the figure and get the colors cycle
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    for col in plt.rcParams['axes.prop_cycle'].by_key()['color']: colors.append(col)  # Triplicate the color cycle
    for col in plt.rcParams['axes.prop_cycle'].by_key()['color']: colors.append(col)  # in case of many eigenvalues
    # Loop over the sorted eigenvalues for ploting the locus and count the encirclements
    for idx in range(eigenvalues_sorted.shape[1]):
        # Plot the eigenvalue locus
        ax[0].plot(x[:,idx], y[:,idx], color=colors[idx], linestyle='solid', linewidth=2.0, label=r'$\lambda_{'+format(idx+1,'.0f')+r'}$')
        ax[0].plot(x[:,idx], -y[:,idx], color=colors[idx], linestyle='solid', linewidth=2.0, label='_nolegend_')
        ax[1].plot(x[:,idx],y[:,idx], color=colors[idx],linestyle='solid',linewidth=2.0,label=r'$\lambda_{' + format(idx+1,'.0f')+r'}$')
        ax[1].plot(x[:,idx], -y[:,idx], color=colors[idx], linestyle='solid', linewidth=2.0, label='_nolegend_')

        # Count the number of (-1, 0j) encirclements by this eigenvalue locus
        cwi = 0  # Initialize the counters
        ccwi = 0
        for j in range(1,eigenvalues_sorted.shape[0]):
            # Only consider clockwise crossings of the imaginary axis beyond (-1,0j)
            if y[j - 1, idx] < 0 < y[j, idx] and (x[j,idx] < -1):  # x[j-1,idx] < -1 or
                # Check that the (-1,0j) is to the right of the line between (x1,y1) and (x2,y2)
                # If the cross product of vectors (x2-x1, y2-y1) and (-1-x1, 0-y1) is < 0, then (-1,0) is to the right
                if (x[j,idx] - x[j-1,idx])*(0 - y[j-1,idx]) - (y[j,idx] - y[j-1,idx])*(-1 - x[j-1,idx]) < 0:
                    cwi += 1
                    if verbose: print("CW crossing around ",round(0.5*(frequencies[j] + frequencies[j-1]),4)," Hz by lambda =",str(idx))
                    fig1, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(6, 7))
                    ax1.plot([x[j-1,idx],x[j,idx]],[y[j-1,idx],y[j,idx]], color='red', linestyle='solid', linewidth=2.0, label='_nolegend_')
                    ax1.scatter(x[j - 1, idx], y[j - 1, idx], color='green', label=str(frequencies[j-1]))
                    ax1.scatter(x[j, idx], y[j, idx], color='blue', label=str(frequencies[j]))
                    ax1.scatter(-1, 0, marker="+", c='black', label=r'$( -1, 0j )$')
                    ax1.legend(loc='upper right', ncol=1)
                    ax1.minorticks_on()
                    ax1.grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
                    ax1.grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
                    # fig1.show()  # Visualize the plot interactively
                    with open(results_folder + '\\' + filename + "_GNC_lambda_"+str(idx)+"_cw_"+str(cwi)+".pickle", 'wb') as f: pickle.dump(fig1, f)
                    plt.close(fig1)
                    # fig1.clf()

            # Only consider counter-clockwise crossings of the imaginary axis beyond (-1,0j)
            elif y[j - 1, idx] > 0 > y[j, idx] and (x[j-1,idx] < -1 or x[j,idx] < -1):
                if (x[j,idx] - x[j-1,idx])*(0 - y[j-1,idx]) - (y[j,idx] - y[j-1,idx])*(-1 - x[j-1,idx]) > 0:
                    # If the cross product of vectors (x2-x1, y2-y1) and (-1-x1, 0-y1) is > 0, then (-1,0) is to the left
                    # if (x[j,idx] - x[j-1,idx])*(0 - y[j-1,idx]) - (y[j,idx] - y[j-1,idx])*(-1 - x[j-1,idx]) > 0:
                    ccwi += 1
                    if verbose: print("CCW crossing around ",round(0.5*(frequencies[j] + frequencies[j-1]),4)," Hz by lambda =",str(idx))
                    fig1, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(6, 7))
                    ax1.plot([x[j-1,idx],x[j,idx]],[y[j-1,idx],y[j,idx]], color='red', linestyle='solid', linewidth=2.0, label='_nolegend_')
                    ax1.scatter(x[j - 1, idx], y[j - 1, idx], color='green', label=str(frequencies[j-1]))
                    ax1.scatter(x[j, idx], y[j, idx], color='blue', label=str(frequencies[j]))
                    ax1.scatter(-1, 0, marker="+", c='black', label=r'$( -1, 0j )$')
                    ax1.legend(loc='upper right', ncol=1)
                    ax1.minorticks_on()
                    ax1.grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
                    ax1.grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
                    # fig1.show()  # Visualize the plot interactively
                    with open(results_folder + '\\' + filename + "_GNC_lambda_"+str(idx)+"_ccw_"+str(ccwi)+".pickle", 'wb') as f: pickle.dump(fig1, f)
                    plt.close(fig1)

        cw.append(cwi)  # Add the counters to the list
        ccw.append(ccwi)
    # print("CC: ",cw,"\nCCW: ",ccw)
    N = sum(cw) - sum(ccw)  # Net number of clockwise encirclements
    if N > 0:
        stable_system = False
        if verbose: print("\n GNC stability assessment: unstable closed-loop system \n")
    elif N < 0:
        stable_system = False
        if verbose: print("\n GNC stability assessment: unstable subsystem \n")
    else:
        stable_system = True
        if verbose: print("\n GNC stability assessment: stable closed-loop system if subsystems are stable \n")

    # Plot the unit circle and the critical point
    th = np.linspace(-np.pi * 1.01, np.pi * 1.01, 314)
    ax[0].plot(np.cos(th), np.sin(th), color='black', linestyle='dotted', linewidth=2.0, label='Unit circle')
    ax[0].scatter(-1, 0, marker="+", c='blue', label=r'$( -1, 0j )$')
    ax[0].minorticks_on()
    ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[0].set_title('Eigenloci between '+format(frequencies[0],'.1f')+' and '+format(frequencies[-1],'.1f') + ' Hz')
    ax[0].set_xlim([np.min(x,axis=None), np.max(x,axis=None)])
    ax[0].set_ylim([np.min(y,axis=None), np.max(y,axis=None)])
    ax[0].set_xlabel('Real axis')
    ax[0].set_ylabel('Imaginary axis')
    ax[0].legend(loc='upper right', ncol=4)

    ax[1].plot(np.cos(th), np.sin(th), color='black', linestyle='dotted', linewidth=2.0, label='Unit circle')
    ax[1].scatter(-1, 0,s=4*rcParams['lines.markersize'] ** 2, marker="+", c='blue', label=r'$( -1, 0j )$')
    ax[1].set_xlim([-2.0, 2.0])
    ax[1].set_ylim([-2.0, 2.0])
    ax[1].minorticks_on()
    ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    # plt.show()  # Visualize the plot interactively
    fig.savefig(results_folder + '\\' + filename + "_GNC.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + filename + "_GNC.pickle", 'wb') as f:
        pickle.dump(fig, f)
    plt.close(fig)

    # Save the eigenloci
    loci = [eigenvalues_sorted[:, idx] for idx in range(eigenvalues_sorted.shape[1])]
    loci.insert(0, frequencies)
    loci = tuple(loci)
    np.savetxt(results_folder + '\\' + filename + '_GNC.txt', np.stack(loci, axis=1), delimiter='\t',
               header="Frequency [Hz]\tEigenvalues", comments='')

    return stable_system

def small_gain(G1, G2, frequencies, results_folder=None, filename=None, variables=None):
    # Applies a conservative version of the small-gain theorem as |G1*G2| <= |G1|*|G2| < 1
    # The loop gain is L = G1*G2
    S1 = np.linalg.svd(G1, compute_uv=False)
    S2 = np.linalg.svd(G2, compute_uv=False)

    if not path.exists(results_folder): makedirs(results_folder)  # Create results folder if it does not exist

    fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(8, 6))
    ax[0].plot(frequencies, 1.0/np.max(S1, axis=1), color='blue', linestyle='solid', linewidth=2.0, label=r"1 / max $\sigma (\mathbf{G}_1)$")
    if variables is not None:
        # Find the position of the block diagonal matrices as they are surrounded by zeros
        indices = []  # Tuple of start and end indeces of each matrix
        start_index = 0
        for index in range(G2.shape[1] - 1):
            if G2[0, index, start_index] == 0:  # If the next element is zero, then boundary of block matrix if defined
                indices.append((start_index, index - 1))  # There is a block matrix between start_index and index
                start_index = index  # Initialize start index of next matrix
        indices.append((start_index, G2.shape[1] - 1))  # Last block matrix

        # Plot the maximum singular values of the different block matrices and the whole matrix
        for block_pos in range(len(indices)):
            start_idx = indices[block_pos][0]
            end_idx = indices[block_pos][1] + 1
            S2_block = np.max(np.linalg.svd(G2[:,start_idx:end_idx,start_idx:end_idx], compute_uv=False), axis=1)
            ax[0].plot(frequencies, S2_block, linestyle='solid',linewidth=2.0, label=", ".join(variables[start_idx:end_idx]))
    ax[0].plot(frequencies, np.max(S2, axis=1), color='red', linestyle='dashed', linewidth=2.0,label=r"max $\sigma (\mathbf{G}_2)$")

    # Setings for upper plot
    ax[0].set_xscale("log")
    ax[0].set_yscale("log")
    ax[0].minorticks_on()
    ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[0].set_ylabel(r'max $\sigma ( \cdot )$')
    ax[0].set_title('Singular value analysis over ' + str(len(frequencies)) + ' frequencies')
    ax[0].set_xlim([frequencies[0], frequencies[-1]])
    # ax[0].set_xlabel('Frequency [Hz]')
    ax[0].legend(loc='best', fancybox=True, shadow=True, ncol=1)

    S12 = np.linalg.svd(np.matmul(G1,G2), compute_uv=False)
    ax[1].plot(frequencies, np.max(S12, axis=1), color='black', linestyle='solid', linewidth=2.0, label=r"max $\sigma (\mathbf{G}_1  \mathbf{G}_2)$")
    ax[1].plot(frequencies, np.multiply(np.max(S1, axis=1), np.max(S2, axis=1)), color='green', linestyle='dashed',
            linewidth=2.0, label=r"max $\sigma (\mathbf{G}_1) \cdot $ max $\sigma (\mathbf{G}_2)$")
    ax[1].plot([frequencies[0], frequencies[-1]],[1, 1], color='grey', linestyle='dotted',linewidth=2.0, label='_nolegend_')
    ax[1].set_xscale("log")
    ax[1].set_yscale("log")
    ax[1].minorticks_on()
    ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[1].set_ylabel('Unitless')
    ax[1].set_xlim([frequencies[0], frequencies[-1]])
    ax[1].set_xlabel('Frequency [Hz]')
    ax[1].legend(loc='best', fancybox=True, shadow=True, ncol=1)
    fig.savefig(results_folder + '\\' + filename + "_gain.pdf", format="pdf", bbox_inches="tight")

    with open(results_folder + '\\' + filename + "_gain.pickle", 'wb') as f:
        pickle.dump(fig, f)
    np.savetxt(results_folder + '\\' + filename + '_gain.txt',
                np.stack((frequencies, np.max(S1, axis=1), np.max(S2, axis=1), np.max(S12, axis=1)), axis=1),
                delimiter='\t', header="f\t" + "max_sigma_G1\t" + "max_sigma_G2\t" + "max_sigma_G1_G2", comments='')
    plt.close(fig)

def EVD(G, frequencies, bus_names=None, results_folder=None, filename=None, verbose=True):
    if bus_names is None: bus_names = [str(bus+1) for bus in range(G.shape[1])]  # Sorted numbers if names not provided
    if not path.exists(results_folder): makedirs(results_folder)  # Create results folder if it does not exist

    # 1) Eigenvalue decomposition over the frequency
    eigenvalues, right_eigenvectors = np.linalg.eig(G)

    # Sorting of eigenvalues for a continuous plot based on minimum changes between adjacent frequencies
    eigenvalues_sorted = np.empty(eigenvalues.shape, dtype='cdouble')  # Initialization of sorted eigenvalues
    eigenvalues_sorted[0, :] = eigenvalues[0, :]  # First frequency as reference for the sorting
    for idx in range(1, eigenvalues.shape[0]):
        eig_1 = eigenvalues_sorted[idx - 1, :]  # Previous eigenvalues
        eig_2 = eigenvalues[idx, :]  # Current eigenvalues
        # Create matrix of distances between eigenvalues
        x, y = np.meshgrid(np.real(eig_1), np.real(eig_2), indexing='ij')
        d_real = np.abs(x - y)
        x, y = np.meshgrid(np.imag(eig_1), np.imag(eig_2), indexing='ij')
        d_imag = np.abs(x - y)
        d_abs = np.sqrt(np.square(d_real) + np.square(d_imag))  # Absolute distance by element-wise operations
        # Solve the linear sum assignment problem to find the minimum variation and thus the correct order
        col_ind = linear_sum_assignment(d_abs)[1]  # The absolute distance is the cost matrix
        eigenvalues_sorted[idx, :] = eigenvalues[idx, col_ind]  # Sort the eigenvalues
        right_eigenvectors[idx, :] = right_eigenvectors[idx][:,col_ind]  # Sort the eigenvectors

    left_eigenvectors = np.linalg.inv(right_eigenvectors)

    # 2) Oscillation modes identification
    lambda_re = np.real(eigenvalues_sorted)  # Real part
    lambda_imag = np.real(eigenvalues_sorted)  # Imaginary part
    lambda_abs = np.abs(eigenvalues_sorted)  # Absolute value (magnitude)

    # 2.1) Oscillation mode identification based on the magnitude peaks of G = closed-loop impedance matrix
    idx_lambda_max = np.argmax(lambda_abs,axis=0)  # Frequency index of the maximum magnitude of each eigenvalue
    idx_lambda_max_max = np.argmax([lambda_abs[idx_lambda_max[idx],idx] for idx in range(eigenvalues.shape[1])])  # Critical mode = the highest mag peak
    freq_idx = idx_lambda_max[idx_lambda_max_max]  # Oscillation frequency index; or also freq_indices[idx_lambda_min]
    if verbose: print("The main oscillation frequency is",round(frequencies[freq_idx],2),"Hz based on the magnitude of eigenvalue",idx_lambda_max_max+1,"=",round(eigenvalues_sorted[idx_lambda_max[idx_lambda_max_max],idx_lambda_max_max], 5))

    # # 2.2) Based on the minimum real part at imaginary zero-crossing (used in the Positive Net Damping criterion)
    # sign_changes = np.diff(np.sign(lambda_imag),axis=0)
    # critical_points = np.nonzero(sign_changes)  # Frequency of sign change from + to - or viceversa of the imaginary parts
    # freq_indices = critical_points[0]
    # lambda_indices = critical_points[1]
    # # Frequency index of the minimum real part of each eigenvalue over the critical frequencies
    # critical_lambdas = [lambda_re[freq_indices[idx], lambda_indices[idx]] for idx in range(len(freq_indices))]  # Real part at the critical frequencies
    # idx_lambda_min = np.argmin(critical_lambdas)
    # print("According to the PND, the critical mode is at",round(frequencies[freq_indices[idx_lambda_min]],3),"Hz based on the minimum real part of eigenvalue",lambda_indices[idx_lambda_min]+1)
    # print(" Eigenvalue",lambda_indices[idx_lambda_min]+1, "=",round(eigenvalues_sorted[freq_indices[idx_lambda_min],lambda_indices[idx_lambda_min]],5))
    # print(" The slope of the imaginary part is",round(lambda_imag[freq_indices[idx_lambda_min],lambda_indices[idx_lambda_min]]-lambda_imag[freq_indices[idx_lambda_min]-1,lambda_indices[idx_lambda_min]],5))
    # freq_idx = freq_indices[idx_lambda_min] # Oscillation frequency index or also round(critical_lambdas[idx_lambda_min],5)

    # 3) Compute the bus participation factors (PFs) of the critical eigenvalue at the oscillation frequency
    # Controllability (right eigenvectors) and observability (transpose of left eigenvectors)
    Obs = right_eigenvectors[freq_idx, :]
    Cont = np.transpose(left_eigenvectors[freq_idx, :])
    # PF[row = bus, column = mode]
    PF = right_eigenvectors[freq_idx, :] * np.transpose(left_eigenvectors[freq_idx, :])  # Element-wise product
    PF_mode = PF[:,idx_lambda_max_max]  # Select the target mode
    # The controllability, observability and PF of the critical mode at each bus
    if verbose: print("Bus"+(max([len(bus) for bus in bus_names])-3)*" "+"\t","Cont.\t","Obs.\t","PF")  # Header
    for idx, bus in enumerate(bus_names):
        if verbose: 
            print(bus+"\t", f"{np.abs(Cont[idx,idx_lambda_max_max]) / np.sum(np.abs(Cont[:,idx_lambda_max_max])):.4f}"+"\t",
              f"{np.abs(Obs[idx,idx_lambda_max_max]) / np.sum(np.abs(Obs[:,idx_lambda_max_max])):.4f}"+"\t",
              f"{np.abs(PF_mode[idx]) / np.sum(np.abs(PF_mode)):.4f}")

    # 4) Plot the eigenvalues over frequency
    fig, ax = plt.subplots(nrows=3, ncols=1, figsize=(6, 8))  # Create the figure and get the colors cycle
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    for col in plt.rcParams['axes.prop_cycle'].by_key()['color']: colors.append(col)  # Triplicate colour cycle
    for col in plt.rcParams['axes.prop_cycle'].by_key()['color']: colors.append(col)  # in case of many eigenvalues
    # Loop over the sorted eigenvalues and plot them over the frequency range
    for idx in range(eigenvalues_sorted.shape[1]):
        # Plot the eigenvalue locus
        ax[0].plot(frequencies, lambda_abs[:,idx], color=colors[idx], linestyle='solid', linewidth=2.0,
                   label=r'$\lambda_{' + format(idx + 1, '.0f') + r'}$')
        ax[1].plot(frequencies, lambda_re[:, idx], color=colors[idx], linestyle='solid', linewidth=2.0,
                   label=r'$\lambda_{' + format(idx + 1, '.0f') + r'}$')
        ax[2].plot(frequencies, lambda_imag[:, idx], color=colors[idx], linestyle='solid', linewidth=2.0,
                   label=r'$\lambda_{' + format(idx + 1, '.0f') + r'}$')

    # Figure settings and save to pdf
    ax[0].minorticks_on()
    ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[0].set_title('Eigenvalue decomposition between ' + format(frequencies[0], '.1f') + ' and ' + format(frequencies[-1], '.1f') + ' Hz')
    ax[0].set_xlim([frequencies[0], frequencies[-1]])
    ax[0].set_ylim([np.min(lambda_abs, axis=None), np.max(lambda_abs, axis=None)])
    ax[0].set_ylabel('Magnitude')
    ax[0].set_xscale("log")
    ax[0].set_yscale("log")
    ax[0].legend(loc='lower right', ncol=4)

    ax[1].minorticks_on()
    ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[1].set_xlim([frequencies[0], frequencies[-1]])
    ax[1].set_ylim([np.min(lambda_re, axis=None), np.max(lambda_re, axis=None)])
    ax[1].set_ylabel('Real part')
    ax[1].set_xscale("log")

    ax[2].minorticks_on()
    ax[2].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[2].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[2].set_xlim([frequencies[0], frequencies[-1]])
    ax[2].set_ylim([np.min(lambda_imag, axis=None), np.max(lambda_imag, axis=None)])
    ax[2].set_xlabel('Frequency [Hz]')
    ax[2].set_ylabel('Imaginary part')
    ax[2].set_xscale("log")

    # plt.show()  # Visualize the plot interactively
    fig.savefig(results_folder + '\\' + filename + "_EVD.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + filename + "_EVD.pickle", 'wb') as f:
        pickle.dump(plt.gcf(), f)
    plt.close(fig)

    # Save the EVD into a text file
    evd_results = [eigenvalues_sorted[:,idx] for idx in range(eigenvalues_sorted.shape[1])]
    evd_results.insert(0, frequencies)
    evd_results = tuple(evd_results)
    np.savetxt(results_folder + '\\' + filename + '_EVD.txt', np.stack(evd_results, axis=1), delimiter='\t',
               header="Frequency [Hz]\t" + "\t".join(bus_names), comments='')

def nyquist_det(L, frequencies, results_folder=None, filename=None, verbose=True, offset=0.0, draw_arrows=True, show_plot=False):
    # Stability assessment based on the determinant of I + L
    # Theorem 4.14 in J. Maciejowski, Multivariable feedback design. Addison-Wesley, 1989, assuming standalone stable subsystems,
    # i.e. L does not have any open-loop unstable poles, the stability conditions are:
    # a) Zero net number of clockwise encirclements of (0,j0) by det[I+L(s)] as s travels the imaginary axis from 0 to +j*infinity
    # b) No crossings of the orgin by by det[I + L(s)] as s travels the imaginary axis from 0 to +j*infinity
    # 
    # Note: the offset parameter can be used to look at other points in the real axis instead of (0,j0), e.g. (-1,j0) as in the eigenvalue loci of L
    if not path.exists(results_folder): makedirs(results_folder)  # Create results folder if it does not exist
    
    det = np.linalg.det(np.identity(2) + L) + offset  
    x = np.real(det)
    y = np.imag(det)

    # Count the number of (offset, 0j) encirclements by det
    cwi = 0  # Initialize the counters
    ccwi = 0
    for j in range(1,len(frequencies)):
        # Only consider clockwise crossings of the imaginary axis beyond (offset,0j)
        if y[j - 1] < 0 < y[j] and (x[j] < offset):  # x[j-1] < offset or
            # Check that the (offset,0j) is to the right of the line between (x1,y1) and (x2,y2)
            # If the cross product of vectors (x2-x1, y2-y1) and (offset-x1, 0-y1) is < 0, then (offset,0) is to the right
            if (x[j] - x[j-1])*(0 - y[j-1]) - (y[j] - y[j-1])*(offset - x[j-1]) < 0:
                cwi += 1
                if verbose: print("CW crossing around ",round(0.5*(frequencies[j] + frequencies[j-1]),4)," Hz")
                fig1, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(6, 7))
                ax1.plot([x[j-1],x[j]],[y[j-1],y[j]], color='red', linestyle='solid', linewidth=2.0, label='_nolegend_')
                ax1.scatter(x[j-1], y[j-1], color='green', label=str(frequencies[j-1]))
                ax1.scatter(x[j], y[j], color='blue', label=str(frequencies[j]))
                ax1.scatter(offset, 0, marker="+", c='black', label=r'$( -1, 0j )$')
                ax1.legend(loc='upper right', ncol=1)
                ax1.minorticks_on()
                ax1.grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
                ax1.grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
                if show_plot: fig1.show()  # Visualize the plot interactively
                with open(results_folder + '\\' + filename + "_det_cw_"+str(cwi)+".pickle", 'wb') as f: pickle.dump(fig1, f)
                plt.close(fig1)

        # Only consider counter-clockwise crossings of the imaginary axis beyond (offset,0j)
        elif y[j - 1] > 0 > y[j] and (x[j-1] < offset or x[j] < offset):
            if (x[j] - x[j-1])*(0 - y[j-1]) - (y[j] - y[j-1])*(offset - x[j-1]) > 0:
                # If the cross product of vectors (x2-x1, y2-y1) and (offset-x1, 0-y1) is > 0, then (offset,0) is to the left
                ccwi += 1
                if verbose: print("CCW crossing around ",round(0.5*(frequencies[j] + frequencies[j-1]),4)," Hz")
                fig1, ax1 = plt.subplots(nrows=1, ncols=1, figsize=(6, 7))
                ax1.plot([x[j-1],x[j]],[y[j-1],y[j]], color='red', linestyle='solid', linewidth=2.0, label='_nolegend_')
                ax1.scatter(x[j-1], y[j-1], color='green', label=str(frequencies[j-1]))
                ax1.scatter(x[j], y[j], color='blue', label=str(frequencies[j]))
                ax1.scatter(offset, 0, marker="+", c='black', label=r'$( -1, 0j )$')
                ax1.legend(loc='upper right', ncol=1)
                ax1.minorticks_on()
                ax1.grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
                ax1.grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
                if show_plot: fig1.show()  # Visualize the plot interactively
                with open(results_folder + '\\' + filename + "_det_ccw_"+str(ccwi)+".pickle", 'wb') as f: pickle.dump(fig1, f)
                plt.close(fig1)

    N = cwi - ccwi  # Net number of clockwise encirclements
    if N > 0:
        stable_system = False
        if verbose: print("\n GNC stability assessment: unstable closed-loop system \n")
    elif N < 0:
        stable_system = False
        if verbose: print("\n GNC stability assessment: unstable subsystem \n")
    else:
        stable_system = True
        if verbose: print("\n GNC stability assessment: stable closed-loop system if subsystems are stable \n")

    fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(6, 7))  # Create the figure and get the colors cycle
    # Plot the critical point
    ax[0].scatter(offset, 0, marker="+", c='blue', label=r'$( '+str(round(offset,0))+', 0j )$')
    ax[0].plot(x, y, color='red', linestyle='solid', linewidth=2.0)
    if draw_arrows:
        da = int(np.log(L.shape[0] + 1))  # decimate the number of arrows
        ax[0].quiver(x[::da], y[::da], np.gradient(x)[::da], np.gradient(y)[::da], angles='xy', scale_units='xy',
                    scale=da, linewidth=1, edgecolor='black', facecolor='green')
    id0 = np.argmin(np.abs(frequencies - 50.0))
    ax[0].text(x[id0], y[id0], str(round(frequencies[id0], 2)) + ' Hz', fontsize=12, color='blue', ha='right', va='bottom')
    ax[0].text(x[-1], y[-1], str(round(frequencies[-1], 2)) + ' Hz', fontsize=12, color='blue', ha='right', va='bottom')
    ax[0].text(x[0], y[0], str(round(frequencies[0], 2)) + ' Hz', fontsize=12, color='blue', ha='right', va='bottom')

    ax[0].minorticks_on()
    ax[0].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[0].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[0].set_title(str(offset)+r' + det[I + L(s)] between '+format(frequencies[0], '.1f')+' and ' + format(frequencies[-1], '.1f') + ' Hz')
    ax[0].set_xlim([np.min(x, axis=None), np.max(x, axis=None)])
    ax[0].set_ylim([np.min(y, axis=None), np.max(y, axis=None)])
    ax[0].set_xlabel('Real axis')
    ax[0].set_ylabel('Imaginary axis')
    ax[0].legend(loc='best', ncol=1)

    ax[1].plot(x, y, color='red', linestyle='solid', linewidth=2.0)
    if draw_arrows:
        ax[1].quiver(x, y, np.gradient(x), np.gradient(y), angles='xy', scale_units='xy',
                    scale=1, linewidth=1, edgecolor='black', facecolor='green')
    ax[1].scatter(offset, 0, s=4 * rcParams['lines.markersize'] ** 2, marker="+", c='blue', label=r'$( ' + str(offset) + ', 0j )$')
    ax[1].set_xlim([-1.0+offset, 1.0+offset])
    ax[1].set_ylim([-1.0+offset, 1.0+offset])
    ax[1].minorticks_on()
    ax[1].grid(visible=True, which='major', color='k', linestyle='-', linewidth=0.5)
    ax[1].grid(visible=True, which='minor', color='tab:gray', alpha=0.5, linestyle='-', linewidth=0.5)
    ax[1].set_xlabel('Real axis')
    ax[1].set_ylabel('Imaginary axis')

    fig.savefig(results_folder + '\\' + filename + "_det.pdf", format="pdf", bbox_inches="tight")
    with open(results_folder + '\\' + filename + "_det.pickle", 'wb') as f:
        pickle.dump(fig, f)
    if show_plot: plt.show()  # Visualize the plot interactively
    plt.close(fig)

    # Save the results
    np.savetxt(results_folder + '\\' + filename + '_det.txt', np.stack((frequencies,det), axis=-1), delimiter='\t',
               header="Frequency [Hz]\t"+str(offset)+"+det[I + L(s)]", comments='')
    
    return stable_system


"""
Function to read the addmittance files as obtained with the tool
"""
"""
Copyright (C) 2024  Francisco Javier Cifuentes Garcia

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['read_admittance']

from os import listdir
import numpy as np  # Numerical python functions

class Admittance:
    def __init__(self, variables, admittance, f, node=False):
        self.variables = variables  # Variable names including the block, side and d,q,dc
        self.vars = []  # Names of the variables without the block side for variable pairing
        self.y = admittance  # Admittance data
        self.f = f  # Frequency data
        # Determine if it is an AC, DC or ACDC matrix and extract block names
        self.blocks_info = {}  # Keys are the block names, contents: side (1 or 2) and type (AC, or DC)
        self.blocks = []
        y_type = []
        for name_loop in variables:
            if "dc" in name_loop.split("_")[-1]:
                y_type.append("DC")
            else:
                y_type.append("AC")
            # Remove the ending: _d, _dc, _q and add it as a dict key
            b = "_".join(name_loop.split("_")[:-1])
            self.vars.append("_".join([b[:-2],name_loop.split("_")[-1]]))
            if b not in self.blocks: self.blocks.append(b)  # "-".join(b.split("-")[:-1])
            if self.blocks[-1] not in list(self.blocks_info.keys()):
                self.blocks_info[self.blocks[-1]] = {"type": y_type[-1], "side": b[-1]}  # b..split("-")[-1]
        y_type = list(set(y_type))
        if len(y_type) != 1: y_type = ["ACDC"]
        # print(self.blocks[-1],self.blocks_info[self.blocks[-1]])  # Only the name of the blocks as they appear in the matrix (.txt file)
        # print(self.vars)
        self.y_type = y_type[0]  # AC, DC or ACDC
        if node or self.y_type == "ACDC" or (admittance.shape[1] == 2 and self.y_type == "AC") or (admittance.shape[1] == 1 and self.y_type == "DC"):
            self.node = True
        else:
            self.node = False

def read_admittance(path=None, involved_blocks=None, file_name=None, file_root=None):
    # involved_blocks is a list of strings containing the names of the blocks with sides included
    # Either admittance_type, involved_blocks and path (and name root) can be specified or full file_name
    if file_name is None:
        if (involved_blocks or path) is None:
            print('\nError: One or more required arguments are missing. \n')
            return
        else:
            # Look for the text file involving the indicated blocks
            file_name = [file for file in listdir(path) if (file.endswith("#.txt") and all(x in file for x in involved_blocks))]
            if file_root is not None: file_name = [file for file in file_name if file.startswith(file_root+"#")]
            # and (file.count("#Y_"+admittance_type+"#") > 0) indicated admittance type, admittance_type=None
            # and (file.count("#") == len(involved_blocks)+2) Just as many blocks as involved
    else:
        if path is None:
            print('\nError: File path is missing. \n')
            return

    # Read the variable names
    with open(path + '\\' + file_name[0], 'r') as f:
        variables = f.readline().strip('\n').split()

    # Load the data
    # print("Loading",path + '\\' + file_name[0])
    data = np.loadtxt(path + '\\' + file_name[0], dtype='cdouble', skiprows=1)
    freq = np.real(data[:, 0])  # Extract frequency column
    data = data[:, 1:]  # Remove frequency column

    return Admittance(variables[1:], data.reshape(data.shape[0], int(np.sqrt(data.shape[1])), int(np.sqrt(data.shape[1]))), freq)


# # Testing code
# pth = r"C:\Users\fcifuent\Desktop\KU Leuven\Projects\Z based analysis\Freq measurement\P2P test system\Tool_test\Results"
# # name = "Ytest#Y_AC#3-AC-2#4-AC-1#.txt"
# # y = read_admittance(path=pth, file_name=name)
# block = ["4-AC-2"]
# y = read_admittance(path=pth, involved_blocks=block)
# print(y.y.shape[1])
# print(y.variables)
# print(y.blocks)
# print(y.blocks_info)
# print(y.node)